# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import gym
import os
from unittest import SkipTest
import numpy as np
import pytest
from . import multigym


GYM_ENV_NAMES = multigym.GymMulti.get_env_names()


def test_multigym() -> None:
    raise SkipTest("Skipping GYM for now.")
    for env_name in multigym.GymMulti.ng_gym:
        assert env_name in GYM_ENV_NAMES, f"{env_name} unknown!"
        assert env_name not in multigym.NO_LENGTH, f"{env_name} in no length and in ng_gym!"
    for env_name in multigym.GUARANTEED_GYM_ENV_NAMES:
        assert env_name in GYM_ENV_NAMES, f"{env_name} should be guaranteed!"
    assert len(GYM_ENV_NAMES) >= 10 or os.name == "nt"


# def test_cartpole() -> None:
#    func = multigym.GymMulti(name="CartPole-v0", control="neural", neural_factor=1, randomized=True)
#    candidate = func.parametrization.sample()
#    results = [func.evaluation_function(candidate) for _ in range(40)]
#    assert min(results) != max(results), "CartPole should not be deterministic."
#
#
# def test_sparse_cartpole() -> None:
#    func = multigym.GymMulti(
#        name="CartPole-v0", control="neural", neural_factor=1, randomized=True, sparse_limit=2
#    )
#    param = func.parametrization.sample()
#    func(*param.args, **param.kwargs)
#    candidate = func.parametrization.sample()
#    results = [func.evaluation_function(candidate) for _ in range(40)]
#    assert min(results) != max(results), "CartPole should not be deterministic."
#
#
@pytest.mark.parametrize("name", ["LunarLander-v2"])  # type: ignore
def test_run_multigym(name: str) -> None:
    raise SkipTest("Skipping GYM for now.")
    if os.name == "nt" or np.random.randint(8) or "CubeCrash" in name:
        raise SkipTest("Skipping Windows and running only 1 out of 8")
    if "ANM" in name:
        raise SkipTest("We skip ANM6Easy and related problems.")

    func = multigym.GymMulti(randomized=False, neural_factor=None)
    x = np.zeros(func.dimension)
    value = func(x)
    np.testing.assert_almost_equal(value, 178.2, decimal=2)
    i = GYM_ENV_NAMES.index(name)
    control = multigym.CONTROLLERS[i % len(multigym.CONTROLLERS)]
    print(f"Working with {control} on {name}.")
    func = multigym.GymMulti(
        name,
        control,
        neural_factor=(None if (control == "linear" or "conformant" in control) else 1),
        randomized=bool(np.random.randint(2)),
    )
    y = func.parametrization.sample()
    func(y.value)
    if "stac" in control and "Acrobat" in name:  # Let's check if the memory works.
        np.testing.assert_almost_equal(func(y.value), 500, decimal=2)
    if "stac" in control and "Pendulum-v0" in name:  # Let's check if the memory works.
        np.testing.assert_almost_equal(func(y.value), 1720.39, decimal=2)


gym.envs.register(
    id="TupleActionSpace-v0", entry_point="nevergrad.functions.gym:TupleActionSpace", max_episode_steps=168
)


# def test_tuple_action_space_random() -> None:
#    func = multigym.GymMulti(name="TupleActionSpace-v0", control="conformant", neural_factor=None)
#    val = ng.optimizers.DiscreteOnePlusOne(func.parametrization, budget=100).minimize(func).value
#    reward = min(func(func.parametrization.sample().value) for _ in range(3))
#    assert reward > -80000  # type: ignore
#    assert reward < 0  # type: ignore
#    assert func(val) < reward  # type: ignore


def test_tuple_action_space_neural() -> None:
    raise SkipTest("Skipping GYM for now.")
    func = multigym.GymMulti(name="TupleActionSpace-v0", control="neural", neural_factor=1)
    results_neural = [func(np.random.normal(size=func.dimension)) for _ in range(10)]
    assert min(results_neural) != max(results_neural)  # type: ignore
    assert all(int(r) == r for r in results_neural)  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Creator: Ryan Kroon, Email: rkroon19@gmail.com
# Accompanying paper: Extending Nevergrad, an Optimisation Platform (in directory)

# Evolving Pacing Strategies for Team Pursuit Track Cycling
# Markus Wagner, Jareth Day, Diora Jordan, Trent Kroeger, Frank Neumann
# Advances in Metaheuristics, Vol. 53, Springer, 2013.
# https://link.springer.com/chapter/10.1007/978-1-4614-6322-1_4
# Java code: https://cs.adelaide.edu.au/~markus/pub/TeamPursuit.zip

import numpy as np
import typing as tp
from .. import base
from nevergrad.parametrization import parameter as p
from .womensteampursuit import womensteampursuit
from .mensteampursuit import mensteampursuit


class Cycling(base.ExperimentFunction):

    """
    Team Pursuit Track Cycling Simulator.

    Parameters
    ----------
    strategy: int
        Refers to Transition strategy or Pacing strategy (or both) of the cyclists; this depends on the strategy length.
        Strategy length can only be 30, 31, 61, 22, 23, 45.
        30: mens transition strategy.
        31: mens pacing strategy.
        61: mens transition and pacing strategy combined.
        22: womens transition strategy.
        23: womens pacing strategy.
        45: womens transition and pacing strategy combined.
    """

    def __init__(self, strategy_index: int = 30) -> None:

        # Preliminary stuff.
        women = strategy_index in [22, 23, 45]
        param_transition = p.TransitionChoice([False, True], repetitions=22 if women else 30)
        init = (400 if women else 550) * np.ones(23 if women else 31)
        gender = "Women" if women else "Men"
        param_pacing = p.Array(init=init, lower=200, upper=1200)
        target_function = team_pursuit_simulation

        # optimising transition strategy
        if strategy_index in (22, 30):
            parameter: tp.Any = param_transition
            parameter.set_name(f"{gender} Transition strategy")
            parameter.set_name("")

        # optimising pacing strategy
        elif strategy_index in (23, 31):
            init = 550 * np.ones(strategy_index)
            parameter = param_pacing
            parameter.set_name(f"{gender} Pacing strategy")
            parameter.set_name("")

        # optimising pacing and transition strategies
        elif strategy_index in (45, 61):
            init = 0.5 * np.ones(strategy_index)  # type: ignore
            parameter = p.Instrumentation(transition=param_transition, pacing=param_pacing)
            parameter.set_name(f"{gender} Pacing and Transition")
            # For some reason the name above does not work...
            # It generates a very long name like
            # "(Wom|M)ens Pacing and Transition:[0.5,...
            parameter.set_name("")

        # error raised if invalid strategy length given
        else:
            raise ValueError("Strategy length must be any of 22, 23, 30, 31, 45, 61")
        super().__init__(target_function, parameter)
        # assert len(self.parametrization.sample().value) == strategy_index, f"{len(self.parametrization.sample().value)} != {strategy_index} (init={init} with len={len(init)})."


def team_pursuit_simulation(x) -> float:

    if len(x) == 2:  # Let us concatenate the instrumentation.
        pacing = x[1]["pacing"]
        transition = x[1]["transition"]
    elif len(x) in (30, 22):
        transition = x
        pacing = [550 if len(x) == 30 else 400] * (len(x) + 1)
    elif len(x) in (31, 23):
        pacing = x
        transition = [True, False] * ((len(x) - 1) // 2)
    else:
        raise ValueError(f"len(x) == {len(x)}")

        # What is this ?
        # for i in range(0, len(pacing_strategy)):
        #    pacing_strategy[i] = 100 * pacing_strategy[i] + 200

    # Create a mensteampursuit oor womensteampursuit object.
    team_pursuit: tp.Any = womensteampursuit() if len(pacing) == 23 else mensteampursuit()
    assert len(pacing) in (23, 31)

    # Simulate event with the default strategies
    result = team_pursuit.simulate(transition, pacing)

    if result.get_finish_time() > 10000:  # in case of inf
        return 10000
    else:
        return float(result.get_finish_time())



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# This module has is based on code and ideas from:
# - Mamadou Aliou Barry
# - Marie-Claire Cambourieux
# - Rémi Pollès
# - Antoine Moreau
# from University Clermont Auvergne, CNRS, SIGMA Clermont, Institut Pascal.
#
# Publications:
# - Aliou Barry, Mamadou; Berthier, Vincent; Wilts, Bodo D.; Cambourieux, Marie-Claire; Pollès, Rémi;
#   Teytaud, Olivier; Centeno, Emmanuel; Biais, Nicolas; Moreau, Antoine (2018)
#   Evolutionary algorithms converge towards evolved biological photonic structures,
#   https://arxiv.org/abs/1808.04689
# - Defrance, J., Lemaître, C., Ajib, R., Benedicto, J., Mallet, E., Pollès, R., Plumey, J.-P.,
#   Mihailovic, M., Centeno, E., Ciracì, C., Smith, D.R. and Moreau, A., 2016.
#   Moosh: A Numerical Swiss Army Knife for the Optics of Multilayers in Octave/Matlab. Journal of Open Research Software, 4(1), p.e13.

import typing as tp
import numpy as np
import nevergrad as ng
from nevergrad.ops import mutations
from . import photonics
from .. import base


def _make_parametrization(
    name: str,
    dimension: int,
    bounding_method: str = "bouncing",
    rolling: bool = False,
    as_tuple: bool = False,
) -> ng.p.Parameter:
    """Creates appropriate parametrization for a Photonics problem

    Parameters
    name: str
        problem name, among bragg, chirped, cf_photosic_realistic, cf_photosic_reference and morpho
    dimension: int
        size of the problem among 16, 40 and 60 (morpho) or 80 (bragg and chirped)
    bounding_method: str
        transform type for the bounding ("arctan", "tanh", "bouncing" or "clipping"see `Array.bounded`)
    as_tuple: bool
        whether we should use a Tuple of Array instead of a 2D-array.

    Returns
    -------
    Instrumentation
        the parametrization for the problem
    """
    if name == "bragg":
        shape = (2, dimension // 2)
        bounds = [(2, 3), (30, 180)]
    elif name == "cf_photosic_realistic":
        shape = (2, dimension // 2)
        bounds = [(1, 9), (30, 180)]
    elif name == "cf_photosic_reference":
        shape = (1, dimension)
        bounds = [(30, 180)]
    elif name == "chirped":
        shape = (1, dimension)
        bounds = [(30, 180)]
    elif name == "morpho":
        shape = (4, dimension // 4)
        bounds = [(0, 300), (0, 600), (30, 600), (0, 300)]
    else:
        raise NotImplementedError(f"Transform for {name} is not implemented")
    divisor = max(2, len(bounds))
    assert not dimension % divisor, f"points length should be a multiple of {divisor}, got {dimension}"
    assert (
        shape[0] * shape[1] == dimension
    ), f"Cannot work with dimension {dimension} for {name}: not divisible by {shape[0]}."
    b_array = np.array(bounds)
    assert b_array.shape[0] == shape[0]  # pylint: disable=unsubscriptable-object
    ones = np.ones((1, int(shape[1])))
    init = np.sum(b_array, axis=1, keepdims=True).dot(ones) / 2  # type: ignore
    if as_tuple:
        instrum = ng.p.Instrumentation(
            *[
                ng.p.Array(init=init[:, i]).set_bounds(
                    b_array[:, 0], b_array[:, 1], method=bounding_method, full_range_sampling=True
                )
                for i in range(init.shape[1])  # type: ignore
            ]
        ).set_name("as_tuple")
        assert instrum.dimension == dimension, instrum
        return instrum
    array = ng.p.Array(init=init)  # type: ignore
    if bounding_method not in ("arctan", "tanh"):
        # sigma must be adapted for clipping and constraint methods
        sigma = ng.p.Array(init=[[10.0]] if name != "bragg" else [[0.03], [10.0]]).set_mutation(exponent=2.0)  # type: ignore
        array.set_mutation(sigma=sigma)
    if rolling:
        mutation = mutations.MutationChoice([mutations.Cauchy(), mutations.Translation(axis=1)])
        array = mutation(array)
    array.set_bounds(b_array[:, [0]], b_array[:, [1]], method=bounding_method, full_range_sampling=True)
    array = ng.ops.mutations.Crossover(axis=1)(array).set_name("")
    assert array.dimension == dimension, f"Unexpected {array} for dimension {dimension}"
    return array


class Photonics(base.ExperimentFunction):
    """Function calling photonics code

    Parameters
    ----------
    name: str
        problem name, among bragg, chirped, cf_photosic_realistic, cf_photosic_reference and morpho
    dimension: int
        size of the problem among 16, 40 and 60 (morpho) or 80 (bragg and chirped)
    transform: str
        transform type for the bounding ("arctan", "tanh", "bouncing" or "clipping", see `Array.bounded`)

    Returns
    -------
    float
        the fitness

    Notes
    -----
    - You will require an Octave installation (with conda: "conda install -c conda-forge octave" then re-source dfconda.sh)
    - Each function requires from around 1 to 5 seconds to compute
    - OMP_NUM_THREADS=1 and OPENBLAS_NUM_THREADS=1 are enforced when spawning Octave because parallelization leads to
      deadlock issues here.

    Credit
    ------
    This module is based on code and ideas from:
    - Mamadou Aliou Barry
    - Marie-Claire Cambourieux
    - Rémi Pollès
    - Antoine Moreau
    from University Clermont Auvergne, CNRS, SIGMA Clermont, Institut Pascal.

    Publications
    ------------
    - Aliou Barry, Mamadou; Berthier, Vincent; Wilts, Bodo D.; Cambourieux, Marie-Claire; Pollès, Rémi;
      Teytaud, Olivier; Centeno, Emmanuel; Biais, Nicolas; Moreau, Antoine (2018)
      Evolutionary algorithms converge towards evolved biological photonic structures,
      https://arxiv.org/abs/1808.04689
    - Defrance, J., Lemaître, C., Ajib, R., Benedicto, J., Mallet, E., Pollès, R., Plumey, J.-P.,
      Mihailovic, M., Centeno, E., Ciracì, C., Smith, D.R. and Moreau, A. (2016)
      Moosh: A Numerical Swiss Army Knife for the Optics of Multilayers in Octave/Matlab. Journal of Open Research Software, 4(1), p.e13.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name: str,
        dimension: int,
        bounding_method: str = "clipping",
        rolling: bool = False,
        as_tuple: bool = False,
    ) -> None:
        assert name in [
            "bragg",
            "morpho",
            "chirped",
            "cf_photosic_reference",
            "cf_photosic_realistic",
        ], f"Unknown {name}"
        self.name = name + ("_as_tuple" if as_tuple else "")
        self._as_tuple = as_tuple
        self._base_func: tp.Callable[[np.ndarray], float] = getattr(photonics, name)
        param = _make_parametrization(
            name=name,
            dimension=dimension,
            bounding_method=bounding_method,
            rolling=rolling,
            as_tuple=as_tuple,
        )
        super().__init__(self._compute, param)

    def to_array(self, *args: tp.Any, **kwargs: tp.Any) -> np.ndarray:
        assert not kwargs
        data = np.concatenate(args).T if self._as_tuple else args[0]
        assert data.size == self.dimension
        return np.array(data, copy=False).ravel()

    def evaluation_function(self, *recommendations: ng.p.Parameter) -> float:
        assert len(recommendations) == 1, "Should not be a pareto set for a singleobjective function"
        recom = recommendations[0]
        x = self.to_array(*recom.args, **recom.kwargs)
        loss = self.function(x)
        assert isinstance(loss, float)
        base.update_leaderboard(f"{self.name},{self.parametrization.dimension}", loss, x, verbose=True)
        return loss

    def _compute(self, *args: tp.Any, **kwargs: tp.Any) -> float:
        x = self.to_array(*args, **kwargs)
        try:
            output = self._base_func(x)
        except Exception:  # pylint: disable=broad-except
            output = float("inf")
        if np.isnan(output):
            output = float("inf")
        return output



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import nevergrad as ng
from .. import base

# pylint: disable=import-outside-toplevel


class PBOFunction(base.ExperimentFunction):
    """
    Pseudo-boolean functions taken from the IOHexperimenter project, adapted for minimization

    Parameters
    ----------
    fid: int
        function number (for a list of functions, please see the
        `documentation <https://www.sciencedirect.com/science/article/abs/pii/S1568494619308099>>`_
    iid: int
        the instance of the function, specifies the transformations in variable and objective space
    dim: int
        The dimensionality of the problem. For function 21 and 23 (N-queens), this needs to be a perfect square
    instrumentation: str
        How the paramerterization is handled. Either "Softmax", in which case the optimizers will work on probabilities
        (so the same parametrization may actually lead to different values), or "Ordered". With this one the optimizer
        will handle a value, which will lead to a 0 if it is negative, and 1 if it is positive.
        This way it is no more stochastic, but the problem becomes very non-continuous

    Notes
    -----
    The IOHexperimenter is part of the IOHprofiler project (IOHprofiler.github.io)

    References
    -----
    For a full description of the available problems and used transformation methods,
    please see "Benchmarking discrete optimization heuristics with IOHprofiler" by C. Doerr et al.
    """

    def __init__(self, fid: int = 1, iid: int = 0, dim: int = 16, instrumentation: str = "Softmax") -> None:
        from IOHexperimenter import IOH_function  # lazy import in case it is not installed

        if fid in [21, 23]:
            assert np.sqrt(
                dim
            ).is_integer(), "Dimension needs to be a perfect square for the selected problem"
        self.f_internal = IOH_function(fid=fid, dim=dim, iid=iid, suite="PBO")
        assert instrumentation in [
            "Softmax",
            "Ordered",
            "Unordered",
        ], "The only valid options for 'instrumentation' are 'Softmax' and 'Ordered'"
        if instrumentation == "Softmax":
            parameterization: ng.p.Parameter = ng.p.Choice([0, 1], repetitions=dim)
        else:
            parameterization = ng.p.TransitionChoice(
                [0, 1], repetitions=dim, ordered=instrumentation == "Ordered"
            )
        super().__init__(self._evaluation_internal, parameterization.set_name(instrumentation))

    def _evaluation_internal(self, x: np.ndarray) -> float:
        assert len(x) == self.f_internal.number_of_variables
        return -float(self.f_internal(x))


class WModelFunction(base.ExperimentFunction):
    """
    Implementation of the W-Model taken from the IOHexperimenter project, adapted for minimization
    Currently supports only OneMax and LeadingOnes as base-functions

    Parameters
    ----------
    base_function: str
        Which base function to use. Either "OneMax" or "LeadingOnes"
    iid: int
        the instance of the function, specifies the transformations in variable and objective space
    dim: int
        The dimensionality of the problem.
    dummy:
        Float between 0 and 1, fractoin of valid bits.
    epistasis:
        size of sub-string for epistasis
    neutrality:
        size of sub-string for neutrality
    ruggedness:
        gamma for ruggedness layper
    instrumentation: str
        How the paramerterization is handled. Either "Softmax", in which case the optimizers will work on probabilities
        (so the same parametrization may actually lead to different values), or "Ordered". With this one the optimizer
        will handle a value, which will lead to a 0 if it is negative, and 1 if it is positive.
        This way it is no more stochastic, but the problem becomes very non-continuous

    Notes
    -----
    The IOHexperimenter is part of the IOHprofiler project (IOHprofiler.github.io)

    References
    -----
    For a description of the W-model in the IOHexperimenter and of the used transformation methods,
    please see "Benchmarking discrete optimization heuristics with IOHprofiler" by C. Doerr et al.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        base_function: str = "OneMax",
        iid: int = 0,
        dim: int = 16,
        dummy: float = 0,
        epistasis: int = 0,
        neutrality: int = 0,
        ruggedness: int = 0,
        instrumentation: str = "Softmax",
    ) -> None:
        from IOHexperimenter import W_model_function  # lazy import in case it is not installed

        assert epistasis <= dim, "Epistasis has to be less or equal to than dimension"
        assert neutrality <= dim, "Neutrality has to be less than or equal to dimension"
        assert ruggedness <= dim**2, "Ruggedness has to be less than or equal to dimension squared"
        assert 0 <= dummy <= 1, "Dummy variable fraction has to be in [0,1]"
        self.f_internal = W_model_function(
            base_function=base_function,
            iid=iid,
            dim=dim,
            dummy=dummy,
            epistasis=epistasis,
            neutrality=neutrality,
            ruggedness=ruggedness,
        )
        assert instrumentation in [
            "Softmax",
            "Ordered",
        ], "The only valid options for 'instrumentation' are 'Softmax' and 'Unordered'"
        if instrumentation == "Softmax":
            parameterization: ng.p.Parameter = ng.p.Choice([0, 1], repetitions=dim)
        else:
            parameterization = ng.p.TransitionChoice([0, 1], repetitions=dim)
        super().__init__(self._evaluation_internal, parameterization.set_name("instrumentation"))

    def _evaluation_internal(self, x: np.ndarray) -> float:
        assert len(x) == self.f_internal.number_of_variables
        assert not set(x) - {0, 1}, f"Pb with input {x} in PBO: not binary."
        return -float(self.f_internal(x))



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# This is based on the specifications of:
# X. Li, K. Tang, M. Omidvar, Z. Yang and K. Qin,
# "Benchmark Functions for the CEC’2013 Special Session and Competition on Large Scale Global Optimization",
# Technical Report, Evolutionary Computation and Machine Learning Group, RMIT University, Australia, 2013.
# and the corresponding source code.

from ._funcs import make_function



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import itertools
import typing as tp
import numpy as np
from nevergrad.common import testing
from . import utils


@testing.parametrized(
    void=({}, ["i1", "i2", "i3"]),
    value=({"c1": "i2-c1"}, ["i2"]),
    function=({"c1": lambda x: x == "i2-c1"}, ["i2"]),
    values=({"c1": ["i3-c1", "i2-c1"]}, ["i2", "i3"]),
    conditions=({"c1": ["i3-c1", "i2-c1"], "c2": "i3-c2"}, ["i3"]),
)
def test_selector(criteria: tp.Any, expected: tp.List[str]) -> None:
    df = utils.Selector(index=["i1", "i2", "i3"], columns=["c1", "c2"])
    for i, c in itertools.product(df.index, df.columns):
        df.loc[i, c] = f"{i}-{c}"
    df_select = df.select(**criteria)
    df_drop = df.select_and_drop(**criteria)
    # indices
    testing.assert_set_equal(df_select.index, expected)
    testing.assert_set_equal(df_drop.index, expected)
    # columns
    testing.assert_set_equal(df_select.columns, df)
    testing.assert_set_equal(df_drop.columns, set(df_select.columns) - set(criteria))
    # values
    for i, c in itertools.product(df_select.index, df_select.columns):
        assert df.loc[i, c] == f"{i}-{c}", "Erroneous values"
    # instance
    assert isinstance(df_select, utils.Selector)
    assert isinstance(df_drop, utils.Selector)


@testing.parametrized(
    multiple=(["c1", "c2"], {(2, 1), (2, 2)}),
    unique=(["c1"], {(2,)}),
    unique_2=(["c2"], {(1,), (2,)}),
    none=([], set()),
)
def test_selector_unique(columns: tp.List[str], expected: tp.Iterable[tp.Tuple[int, int]]) -> None:
    df = utils.Selector(index=["i1", "i2", "i3"], columns=["c1", "c2"], data=[[2, 1], [2, 2], [2, 1]])
    testing.printed_assert_equal(df.unique(columns), expected)


def test_selector_assert_equivalent() -> None:
    select1 = utils.Selector(columns=["a", "b"], data=[[0, 1], [2, 3]])
    select2 = utils.Selector(columns=["b", "a"], data=[[3, 2], [1, 0]])
    select3 = utils.Selector(columns=["a", "b"], data=[[0, 5], [2, 3]])
    select1.assert_equivalent(select2)
    np.testing.assert_raises(AssertionError, select1.assert_equivalent, select3)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
import nevergrad as ng
from . import core


def test_impedence_pix() -> None:
    ep0 = 1
    epf = 9
    x = np.array([3.0, 5.0, 1.0, 9.0])
    dpix = 37
    lam = 400
    output = core.impedance_pix(x, dpix, lam, ep0, epf)
    np.testing.assert_almost_equal(output, 46.64, decimal=2)


def test_arcoating_reflexion_function() -> None:
    func = core.ARCoating(nbslab=4)
    output = func._get_minimum_average_reflexion(np.array([4.56386701, 5.65210553, 6.24006888, 7.18400555]))
    # np.testing.assert_almost_equal(output, 13.320815699203614)  # Before change
    np.testing.assert_almost_equal(output, 12.702, decimal=3)


def test_arcoating_recombination() -> None:
    func = core.ARCoating(nbslab=6)
    func.parametrization.random_state.seed(24)
    arrays: tp.List[ng.p.Array] = []
    for num in [3, 5]:
        arrays.append(func.parametrization.spawn_child())  # type: ignore
        arrays[-1].value = num * np.ones(arrays[0].value.shape)
    arrays[0].recombine(arrays[1])
    expected = [3.0, 3.0, 3.0, 5.0, 3.0, 3.0]
    np.testing.assert_array_equal(arrays[0].value, expected)


def test_arcoating_transform_and_call() -> None:
    nbslab = 8
    func = core.ARCoating(nbslab=nbslab)
    x = 7 * np.random.normal(size=nbslab)  # make sure it touches space boundaries
    data = func.parametrization.spawn_child().set_standardized_data(x).args[0]
    value = func(data)  # should not touch boundaries, so value should be < np.inf
    assert isinstance(value, float)
    assert value < np.inf
    param = func.parametrization.spawn_child().set_standardized_data(np.arange(8))
    np.testing.assert_almost_equal(func(param.value), 16.5538936)
    np.testing.assert_almost_equal(func.evaluation_function(param), 16.5538936)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from . import ac


def test_ac() -> None:
    func = ac.NgAquacrop(4, 12.0)
    x = 50.0 * np.random.rand(func.dimension)
    value = func(x)
    value2 = func(x)
    x = 50.0 * np.random.rand(func.dimension)
    value3 = func(x)
    np.testing.assert_almost_equal(value, value2)
    assert value != value3



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
import functools
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import errors
from . import _layering
from ._layering import Int as Int
from . import data as _data
from .data import Data
from .core import Parameter
from . import discretization
from . import transforms as trans
from . import utils


D = tp.TypeVar("D", bound=Data)
Op = tp.TypeVar("Op", bound="Operation")
BL = tp.TypeVar("BL", bound="BoundLayer")


class Operation(_layering.Layered, _layering.Filterable):

    _LAYER_LEVEL = _layering.Level.OPERATION
    _LEGACY = False

    def __init__(self, *args: tp.Any, **kwargs: tp.Any) -> None:
        super().__init__()
        if any(isinstance(x, Parameter) for x in args + tuple(kwargs.values())):
            raise errors.NevergradTypeError("Operation with Parameter instances are not supported")


class BoundLayer(Operation):

    _LAYER_LEVEL = _layering.Level.OPERATION

    def __init__(
        self,
        lower: tp.BoundValue = None,
        upper: tp.BoundValue = None,
        uniform_sampling: tp.Optional[bool] = None,
    ) -> None:
        """Bounds all real values into [lower, upper]

        Parameters
        ----------
        lower: float or None
            minimum value
        upper: float or None
            maximum value
        method: str
            One of the following choices:
        uniform_sampling: Optional bool
            Changes the default behavior of the "sample" method (aka creating a child and mutating it from the current instance)
            or the sampling optimizers, to creating a child with a value sampled uniformly (or log-uniformly) within
            the while range of the bounds. The "sample" method is used by some algorithms to create an initial population.
            This is activated by default if both bounds are provided.
        """  # TODO improve description of methods
        super().__init__(lower, upper, uniform_sampling)
        self.bounds: tp.Tuple[tp.Optional[np.ndarray], tp.Optional[np.ndarray]] = tuple(  # type: ignore
            None if a is None else trans.bound_to_array(a) for a in (lower, upper)
        )
        both_bounds = all(b is not None for b in self.bounds)
        self.uniform_sampling: bool = uniform_sampling  # type: ignore
        if uniform_sampling is None:
            self.uniform_sampling = both_bounds
        if self.uniform_sampling and not both_bounds:
            raise errors.NevergradValueError("Cannot use full range sampling if both bounds are not set")
        if not (lower is None or upper is None):
            if (self.bounds[0] >= self.bounds[1]).any():  # type: ignore
                raise errors.NevergradValueError(
                    f"Lower bounds {lower} should be strictly smaller than upper bounds {upper}"
                )

    def _normalizer(self) -> trans.Transform:
        if any(b is None for b in self.bounds):
            raise RuntimeError("Cannot use normalized value for not-fully bounded Parameter")
        return trans.Affine(self.bounds[1] - self.bounds[0], self.bounds[0]).reverted()  # type: ignore

    def __call__(self, data: D, inplace: bool = False) -> D:
        """Creates a new Data instance with bounds"""
        new = data if inplace else data.copy()
        # if not utils.BoundChecker(*self.bounds)(new.value):
        #     raise errors.NevergradValueError("Current value is not within bounds, please update it first")
        value = new.value
        new.add_layer(self.copy())
        try:
            new.value = value
        except ValueError as e:
            raise errors.NevergradValueError(
                "Current value is not within bounds, please update it first"
            ) from e
        if all(x is not None for x in self.bounds):
            tests = [data.copy() for _ in range(2)]  # TODO make it simpler and more efficient?
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                for test, bound in zip(tests, self.bounds):
                    val = bound * np.ones(value.shape) if isinstance(value, np.ndarray) else bound[0]  # type: ignore
                    test.value = val
            state = tests[0].get_standardized_data(reference=tests[1])
            min_dist = np.min(np.abs(state))
            if min_dist < 3.0:
                warnings.warn(
                    f"Bounds are {min_dist} sigma away from each other at the closest, "
                    "you should aim for at least 3 for better quality.",
                    errors.NevergradRuntimeWarning,
                )
        return new

    def _layered_sample(self) -> Data:
        if not self.uniform_sampling:
            return super()._layered_sample()  # type: ignore
        root = self._layers[0]
        if not isinstance(root, Data):
            raise errors.NevergradTypeError(f"BoundLayer {self} on a non-Data root {root}")
        shape = super()._layered_get_value().shape
        child = root.spawn_child()
        # send new val to the layer under this one for the child
        new_val = self.random_state.uniform(size=shape)
        del child.value  # make sure there is no cache
        child._layers[self._layer_index].set_normalized_value(new_val)  # type: ignore
        return child

    def set_normalized_value(self, value: np.ndarray) -> None:
        """Sets a value normalized between 0 and 1"""
        new_val = self._normalizer().backward(value)
        self._layers[self._layer_index]._layered_set_value(new_val)

    def get_normalized_value(self) -> np.ndarray:
        """Gets a value normalized between 0 and 1"""
        value = self._layers[self._layer_index]._layered_get_value()
        return self._normalizer().forward(value)

    def _check(self, value: np.ndarray) -> None:
        if not utils.BoundChecker(*self.bounds)(value):
            raise errors.NevergradValueError("New value does not comply with bounds")


class Modulo(BoundLayer):
    """Applies a modulo operation on the array
    CAUTION: WIP
    """

    def __init__(self, module: tp.Any) -> None:
        super().__init__(lower=0, upper=module)
        if not isinstance(module, (np.ndarray, np.float_, np.int_, float, int)):
            raise TypeError(f"Unsupported type {type(module)} for module")
        self._module = module

    def _layered_get_value(self) -> np.ndarray:
        return super()._layered_get_value() % self._module  # type: ignore

    def _layered_set_value(self, value: np.ndarray) -> None:
        self._check(value)
        current = super()._layered_get_value()
        super()._layered_set_value(current - (current % self._module) + value)


class ForwardableOperation(Operation):
    """Operation with simple forward and backward methods
    (simplifies chaining)
    """

    def _layered_get_value(self) -> np.ndarray:
        return self.forward(super()._layered_get_value())  # type: ignore

    def _layered_set_value(self, value: np.ndarray) -> None:
        super()._layered_set_value(self.backward(value))

    def forward(self, value: tp.Any) -> tp.Any:
        raise NotImplementedError

    def backward(self, value: tp.Any) -> tp.Any:
        raise NotImplementedError


class Exponent(ForwardableOperation):
    """Applies an array as exponent of a float"""

    def __init__(self, base: float) -> None:
        super().__init__(base)
        if base <= 0:
            raise errors.NevergradValueError("Exponent must be strictly positive")
        self._base = base
        self._name = f"exp={base:.2f}"

    def forward(self, value: tp.Any) -> tp.Any:
        return self._base**value

    def backward(self, value: tp.Any) -> tp.Any:
        return np.log(value) / np.log(self._base)


class Power(ForwardableOperation):
    """Applies a float as exponent of a Data parameter"""

    def __init__(self, power: float) -> None:
        super().__init__(power)
        self._power = power

    def forward(self, value: tp.Any) -> tp.Any:
        return value**self._power

    def backward(self, value: tp.Any) -> tp.Any:
        return value ** (1.0 / self._power)


class Add(ForwardableOperation):
    """Applies an array as exponent of a floar"""

    def __init__(self, offset: tp.Any) -> None:
        super().__init__(offset)
        self._offset = offset

    def forward(self, value: tp.Any) -> tp.Any:
        return self._offset + value

    def backward(self, value: tp.Any) -> tp.Any:
        return value - self._offset


class Multiply(ForwardableOperation):
    """Applies an array as exponent of a floar"""

    def __init__(self, value: tp.Any) -> None:
        super().__init__(value)
        self._mult = value
        self.name = f"Mult({value})"

    def forward(self, value: tp.Any) -> tp.Any:
        return self._mult * value

    def backward(self, value: tp.Any) -> tp.Any:
        return value / self._mult


class Bound(BoundLayer):
    def __init__(
        self,
        lower: tp.BoundValue = None,
        upper: tp.BoundValue = None,
        method: str = "bouncing",
        uniform_sampling: tp.Optional[bool] = None,
    ) -> None:
        """Bounds all real values into [lower, upper] using a provided method

        See Parameter.set_bounds
        """
        super().__init__(lower=lower, upper=upper, uniform_sampling=uniform_sampling)
        # update instance
        transforms = dict(
            clipping=trans.Clipping,
            arctan=trans.ArctanBound,
            tanh=trans.TanhBound,
            gaussian=trans.CumulativeDensity,
        )
        transforms["bouncing"] = functools.partial(trans.Clipping, bounce=True)  # type: ignore
        if method not in transforms:
            raise errors.NevergradValueError(
                f"Unknown method {method}, available are: {transforms.keys()}\nSee docstring for more help."
            )
        self._method = method
        self._transform = transforms[method](*self.bounds)
        self.set_name(self._transform.name)

    def _layered_get_value(self) -> np.ndarray:
        deep_value = super()._layered_get_value()
        value = self._transform.forward(deep_value)
        if deep_value is not value and self._method in ("clipping", "bouncing"):  # refresh if need be
            # resetting
            super()._layered_set_value(value)
        return value  # type: ignore

    def _layered_set_value(self, value: np.ndarray) -> None:
        super()._layered_set_value(self._transform.backward(value))


class SoftmaxSampling(Int):
    def __init__(self, arity: int, deterministic: bool = False) -> None:
        super().__init__(deterministic=deterministic)
        self.arity = arity
        self.ordered = False

    def _get_name(self) -> str:
        tag = "{det}" if self.deterministic else ""
        return self.__class__.__name__ + tag

    def _layered_get_value(self) -> tp.Any:
        if self._cache is None:
            value = _layering.Layered._layered_get_value(self)
            if value.ndim != 2 or value.shape[1] != self.arity:
                raise ValueError(f"Dimension 1 should be the arity {self.arity}")
            encoder = discretization.Encoder(value, rng=self.random_state)
            self._cache = encoder.encode(deterministic=self.deterministic)
        return self._cache

    def _layered_set_value(self, value: tp.Any) -> tp.Any:
        if not isinstance(value, np.ndarray) and not value.dtype == int:
            raise TypeError("Expected an integer array, got {value}")
        if self.arity is None:
            raise RuntimeError("Arity is not initialized")
        self._cache = value
        out = np.zeros((value.size, self.arity), dtype=float)
        coeff = discretization.weight_for_reset(self.arity)
        out[np.arange(value.size, dtype=int), value] = coeff
        super()._layered_set_value(out)


class AngleOp(Operation):
    """Converts to and from angles from -pi to pi"""

    def _layered_get_value(self) -> tp.Any:
        x = super()._layered_get_value()
        if x.shape[0] != 2:
            raise ValueError(f"First dimension should be 2, got {x.shape}")
        return np.angle(x[0, ...] + 1j * x[1, ...])

    def _layered_set_value(self, value: tp.Any) -> None:
        out = np.stack([fn(value) for fn in (np.cos, np.sin)], axis=0)  # type: ignore
        super()._layered_set_value(out)


def Angles(
    init: tp.Optional[tp.ArrayLike] = None,
    shape: tp.Optional[tp.Sequence[int]] = None,
    deg: bool = False,
    bound_method: tp.Optional[str] = None,
) -> _data.Array:
    """Creates an Array parameter representing an angle from -pi to pi (deg=False)
    or -180 to 180 (deg=True).
    Internally, this keeps track of coordinates which are transformed to an angle.

    Parameters
    ----------
    init: array-like or None
        initial values if provided (either shape or init is required)
    shape: sequence of int or None
        shape of the angle array, if provided (either shape or init is required)
    deg: bool
        whether to return the result in degrees instead of radians
    bound_method: optional str
        adds a bound in the standardized domain, to make sure the values do not
        diverge too much (experimental, the impact is not clear)

    Returns
    -------
    Array
        An Array Parameter instance which represents an angle between -pi and pi

    Notes
    ------
    This API is experimental and will probably evolve in the near future.
    """
    if sum(x is None for x in (init, shape)) != 1:
        raise ValueError("Exactly 1 of init or shape must be provided")
    out_shape = tuple(shape) if shape is not None else np.array(init).shape
    ang = _data.Array(shape=(2,) + out_shape)
    if bound_method is not None:
        Bound(-2, 2, method=bound_method)(ang, inplace=True)
    ang.add_layer(AngleOp())
    with warnings.catch_warnings():  # ignore bounding warning which is irrelevant here
        warnings.simplefilter("ignore", category=errors.NevergradRuntimeWarning)
        Bound(-np.pi, np.pi)(ang, inplace=True)
    if deg:
        ang = ang * (180 / np.pi)
    if init is not None:
        ang.value = init
    return ang



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
from functools import partial
import numpy as np
import sklearn.datasets
from sklearn.tree import DecisionTreeRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.model_selection import train_test_split, KFold
from sklearn.metrics import mean_squared_error
import pandas as pd

from nevergrad.parametrization import parameter as p
from ..base import ExperimentFunction


# pylint: disable=too-many-instance-attributes,too-many-locals,too-many-arguments,too-many-statements
class MLTuning(ExperimentFunction):
    """Class for generating ML hyperparameter tuning problems.
    We propose different possible regressors and different dimensionalities.
    In each case, Nevergrad will optimize the parameters of a scikit learning.

    Parameters
    ----------
    regressor: str
        type of function we can use for doing the regression. Can be "mlp", "decision_tree", "decision_tree_depth",
        "keras_dense_nn", "any".
        "any" means that the regressor has one more parameter which is a discrete choice among sklearn possibilities.
    data_dimension: int
        dimension of the data we generate. None if not an artificial dataset.
    dataset: str
        type of dataset; can be diabetes, boston, artificial, artificialcoos, artificialsquare.
    overfitter: bool
        if we want the evaluation to be the same as during the optimization run. This means that instead
        of train/valid/error, we have train/valid/valid. This is for research purpose, when we want to check if an algorithm
        is particularly good or particularly bad because it fails to minimize the validation loss or because it overfits.

    """

    # Example of ML problem.
    def _ml_parametrization(
        self,
        depth: int,  # Parameters for regression trees.
        criterion: str,
        min_samples_split: float,
        solver: str,  # Parameters for neural nets.
        activation: str,
        alpha: float,
        learning_rate: str,
        regressor: str,  # Choice of learner.
        noise_free: bool,  # Whether we work on the test set (the real cost) on an approximation (CV error on train). -> not really noise
    ) -> float:
        if not self.X_train.size:  # lazzy initialization
            self.make_dataset(self.data_dimension, self.dataset)
        # num_cv-folds cross-validation
        result = 0.0
        # Fit regression model
        if regressor == "decision_tree":
            regr = DecisionTreeRegressor(
                max_depth=depth, criterion=criterion, min_samples_split=min_samples_split, random_state=0
            )
        elif regressor == "mlp":
            regr = MLPRegressor(
                alpha=alpha, activation=activation, solver=solver, learning_rate=learning_rate, random_state=0
            )
        elif regressor == "keras_dense_nn":
            try:
                from tensorflow import keras  # pylint: disable=import-outside-toplevel
            except ImportError as e:
                raise ImportError("Please install keras (pip install keras) to use keras ml tuning") from e

            regr = keras.Sequential(
                [
                    keras.layers.Dense(64, activation=activation, input_shape=(self.X_train.shape[1],)),
                    keras.layers.Dense(1),
                ]
            )
            regr.compile(optimizer=solver, loss="mse", metrics=["mae"])
        else:
            raise ValueError(f"Unknown regressor {regressor}.")

        fit_additional_params = dict(verbose=0, epochs=350) if regressor == "keras_dense_nn" else {}

        if noise_free:  # noise_free is True when we want the result on the test set.
            X_test = self.X_test
            y_test = self.y_test
            regr.fit(self.X_train, self.y_train, **fit_additional_params)
            pred_test = regr.predict(self.X_test)
            return mean_squared_error(self.y_test, pred_test)

        # We do a cross-validation.
        for X, y, X_test, y_test in zip(self.X_train_cv, self.y_train_cv, self.X_valid_cv, self.y_valid_cv):
            assert isinstance(depth, int), f"depth has class {type(depth)} and value {depth}."

            regr.fit(X, y, **fit_additional_params)
            # Predict
            pred_test = regr.predict(X_test)
            try:
                result += mean_squared_error(y_test, pred_test)
            except ValueError:
                result += 5.0e20

        return result / self._cross_val_num  # We return a num_cv-fold validation error.

    def __init__(
        self,
        regressor: str,
        data_dimension: tp.Optional[int] = None,
        dataset: str = "artificial",
        overfitter: bool = False,
    ) -> None:
        self.regressor = regressor
        self.data_dimension = data_dimension
        self.dataset = dataset
        self.overfitter = overfitter
        self.name = regressor + f"Dim{data_dimension}"
        self.num_data = 120  # default for artificial function
        self._cross_val_num = 10  # number of cross validation
        # Dimension does not make sense if we use a real world dataset.
        assert bool("artificial" in dataset) == bool(data_dimension is not None)

        # # Variables for storing the training set and the test set.
        self.X_train: np.ndarray = np.array([])
        self.y_train: np.ndarray
        self.X_test: np.ndarray
        self.y_test: np.ndarray
        # self.X: np.ndarray = np.array([])
        # self.y: np.ndarray

        # Variables for storing the cross-validation splits.
        self.X_train_cv: tp.List[tp.Any] = []  # This will be the list of training subsets.
        self.X_valid_cv: tp.List[tp.Any] = []  # This will be the list of validation subsets.
        self.y_train_cv: tp.List[tp.Any] = []
        self.y_valid_cv: tp.List[tp.Any] = []

        evalparams: tp.Dict[str, tp.Any] = {}
        if regressor == "decision_tree_depth":
            # Only the depth, as an evaluation.
            parametrization = p.Instrumentation(depth=p.Scalar(lower=1, upper=1200).set_integer_casting())
            # We optimize only the depth, so we fix all other parameters than the depth
            params = dict(
                noise_free=False,
                criterion="friedman_mse",
                min_samples_split=0.00001,
                regressor="decision_tree",
                alpha=1.0,
                learning_rate="no",
                activation="no",
                solver="no",
            )
        elif regressor == "any":
            # First we define the list of parameters in the optimization
            parametrization = p.Instrumentation(
                depth=p.Scalar(
                    lower=1, upper=1200
                ).set_integer_casting(),  # Depth, in case we use a decision tree.
                criterion=p.Choice(
                    ["friedman_mse", "poisson", "absolute_error", "squared_error"],
                    # ["mse", "friedman_mse", "mae"]
                ),  # Criterion for building the decision tree.
                min_samples_split=p.Log(
                    lower=0.0000001, upper=1
                ),  # Min ratio of samples in a node for splitting.
                regressor=p.Choice(["mlp", "decision_tree"]),  # Type of regressor.
                activation=p.Choice(
                    ["identity", "logistic", "tanh", "relu"]
                ),  # Activation function, in case we use a net.
                solver=p.Choice(["lbfgs", "sgd", "adam"]),  # Numerical optimizer.
                learning_rate=p.Choice(["constant", "invscaling", "adaptive"]),  # Learning rate schedule.
                alpha=p.Log(lower=0.0000001, upper=1.0),  # Complexity penalization.
            )
            # noise_free is False (meaning that we consider the cross-validation loss) during the optimization.
            params = dict(noise_free=False)
        elif regressor == "decision_tree":
            # We specify below the list of hyperparameters for the decision trees.
            parametrization = p.Instrumentation(
                depth=p.Scalar(lower=1, upper=1200).set_integer_casting(),
                criterion=p.Choice(["friedman_mse", "poisson", "absolute_error", "squared_error"]),
                # criterion=p.Choice(["mse", "friedman_mse", "mae"]),
                min_samples_split=p.Log(lower=0.0000001, upper=1),
                regressor="decision_tree",
            )
            params = dict(
                noise_free=False,
                alpha=1.0,
                learning_rate="no",
                regressor="decision_tree",
                activation="no",
                solver="no",
            )
            evalparams = dict(params, criterion="friedman_mse", min_samples_split=0.00001)
        elif regressor == "mlp":
            # Let us define the parameters of the neural network.
            parametrization = p.Instrumentation(
                activation=p.Choice(["identity", "logistic", "tanh", "relu"]),
                solver=p.Choice(["lbfgs", "sgd", "adam"]),
                regressor="mlp",
                learning_rate=p.Choice(["constant", "invscaling", "adaptive"]),
                alpha=p.Log(lower=0.0000001, upper=1.0),
            )
            params = dict(noise_free=False, regressor="mlp", depth=-3, criterion="no", min_samples_split=0.1)
        elif regressor == "keras_dense_nn":
            parametrization = p.Instrumentation(
                activation=p.Choice(["selu", "sigmoid", "tanh", "relu"]),
                solver=p.Choice(["Adadelta", "RMSprop", "adam"]),
                regressor="keras_dense_nn",
                # metrics=p.Choice(["mae", "mse"]),
            )
            params = dict(
                noise_free=False,
                regressor="keras_dense_nn",
                depth=-3,
                criterion="no",
                min_samples_split=0.1,
                alpha=0.1,
                learning_rate="constant",
            )
        else:
            assert False, f"Problem type {regressor} undefined!"
        # build eval params if not specified
        if not evalparams:
            evalparams = dict(params)
        # For the evaluation we remove the noise (unless overfitter)
        evalparams["noise_free"] = not overfitter
        parametrization.function.proxy = not overfitter
        super().__init__(partial(self._ml_parametrization, **params), parametrization.set_name(""))
        self._evalparams = evalparams

    def evaluation_function(self, *recommendations: p.Parameter) -> float:
        assert len(recommendations) == 1, "Should not be a pareto set for a singleobjective function"
        assert not recommendations[0].args
        kwargs = dict(recommendations[0].kwargs)
        # override with eval parameters (with partial, the eval parameters would be overriden by kwargs)
        kwargs.update(self._evalparams)
        return self._ml_parametrization(**kwargs)

    def make_dataset(self, data_dimension: tp.Optional[int], dataset: str) -> None:
        # Filling datasets.
        rng = self.parametrization.random_state
        if not dataset.startswith("artificial"):
            assert dataset in ["diabetes", "kerasBoston", "auto-mpg", "red-wine", "white-wine"]
            assert data_dimension is None
            sets_url = {
                # "auto-mpg": "http://www-lisic.univ-littoral.fr/~teytaud/files/Cours/Apprentissage/data/auto-mpg.data",
                # "red-wine": "http://www-lisic.univ-littoral.fr/~teytaud/files/Cours/Apprentissage/data/winequality-red.csv",
                # "white-wine": "http://www-lisic.univ-littoral.fr/~teytaud/files/Cours/Apprentissage/data/winequality-white.csv",
                # "auto-mpg": "https://github.com/plotly/datasets/blob/master/auto-mpg.csv",
                # "red-wine": "https://github.com/plotly/datasets/blob/master/winequality-red.csv",
                # "white-wine": "https://github.com/stedy/Machine-Learning-with-R-datasets/blob/master/winequality-white.csv",
                "auto-mpg": "https://raw.githubusercontent.com/plotly/datasets/master/auto-mpg.csv",
                "red-wine": "https://raw.githubusercontent.com/plotly/datasets/master/winequality-red.csv",
                "white-wine": "https://raw.githubusercontent.com/stedy/Machine-Learning-with-R-datasets/master/winequality-white.csv",
            }
            sets_tag = {"auto-mpg": "mpg", "red-wine": "quality", "white-wine": "quality"}
            if dataset == "kerasBoston":
                try:
                    from tensorflow import keras  # pylint: disable=import-outside-toplevel
                except ImportError as e:
                    raise ImportError(
                        "Please install keras (pip install keras) to use keras ml tuning"
                    ) from e

                data = keras.datasets.boston_housing
            elif dataset in sets_tag:
                try:
                    data = pd.read_csv(sets_url[dataset])
                except Exception as e:
                    assert False, f"failing with error {e} for dataset {dataset}"
            else:
                data = {"diabetes": sklearn.datasets.load_diabetes,}[
                    # data = {"boston": sklearn.datasets.load_boston, "diabetes": sklearn.datasets.load_diabetes,}[
                    dataset
                ](return_X_y=True)

            # Half the dataset for training.
            test_ratio = 0.5
            if dataset == "kerasBoston":
                (self.X_train, self.y_train), (self.X_test, self.y_test) = data.load_data(
                    test_split=test_ratio, seed=42
                )
            elif dataset in sets_url:
                if dataset == "auto-mpg":
                    data.drop("name", 1, inplace=True)
                train, test = train_test_split(data, test_size=test_ratio)
                self.y_train = train[sets_tag[dataset]].to_numpy()
                self.y_test = test[sets_tag[dataset]].to_numpy()
                del train[sets_tag[dataset]]
                del test[sets_tag[dataset]]
                self.X_train = train.to_numpy()
                self.X_test = test.to_numpy()
            else:
                rng.shuffle(data[0].T)  # We randomly shuffle the columns.
                self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
                    data[0], data[1], test_size=test_ratio, random_state=42
                )

            num_train_data = len(self.X_train)
            self.num_data = num_train_data

            kf = KFold(n_splits=self._cross_val_num)
            kf.get_n_splits(self.X_train)

            for train_index, valid_index in kf.split(self.X_train):
                self.X_train_cv += [self.X_train[train_index]]
                self.y_train_cv += [self.y_train[train_index]]
                self.X_valid_cv += [self.X_train[valid_index]]
                self.y_valid_cv += [self.y_train[valid_index]]
            return

        assert data_dimension is not None, f"Pb with {dataset} in dimension {data_dimension}"

        # Training set.
        X = np.arange(0.0, 1.0, 1.0 / (self.num_data * data_dimension))
        X = X.reshape(-1, data_dimension)
        rng.shuffle(X)

        target_function: tp.Callable[[np.ndarray], np.ndarray] = {  # type: ignore
            "artificial": np.sin,
            "artificialcos": np.cos,
            "artificialsquare": np.square,
        }[dataset]
        y = np.sum(np.sin(X), axis=1).ravel()
        self.X_train = X  # Training set.
        self.y_train = y  # Labels of the training set.

        # We generate the cross-validation subsets.
        for cv in range(self._cross_val_num):

            # Training set.
            X_train_cv = X[np.arange(self.num_data) % self._cross_val_num != cv].copy()
            y_train_cv = np.sum(target_function(X_train_cv), axis=1).ravel()
            self.X_train_cv += [X_train_cv]
            self.y_train_cv += [y_train_cv]

            # Validation set or test set (noise_free is True for test set).
            X_valid_cv = X[np.arange(self.num_data) % self._cross_val_num == cv].copy()
            X_valid_cv = X_valid_cv.reshape(-1, data_dimension)
            y_valid_cv = np.sum(target_function(X_valid_cv), axis=1).ravel()
            self.X_valid_cv += [X_valid_cv]
            self.y_valid_cv += [y_valid_cv]

        # We also generate the test set.
        X_test = np.arange(0.0, 1.0, 1.0 / 60000)
        rng.shuffle(X_test)
        X_test = X_test.reshape(-1, data_dimension)
        y_test = np.sum(target_function(X_test), axis=1).ravel()
        self.X_test = X_test
        self.y_test = y_test



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# This is based on the specifications of:
# X. Li, K. Tang, M. Omidvar, Z. Yang and K. Qin,
# "Benchmark Functions for the CEC’2013 Special Session and Competition on Large Scale Global Optimization",
# Technical Report, Evolutionary Computation and Machine Learning Group, RMIT University, Australia, 2013.
# and the corresponding source code.
from typing import List, Callable, Optional
from math import exp, sqrt
import numpy as np


def sphere(x: np.ndarray) -> float:
    assert x.ndim == 1
    return float(x.dot(x))


class Elliptic:
    def __init__(self, dimension: int) -> None:
        self.weights = 10 ** np.linspace(0, 6, dimension)  # precompute for speed

    def __call__(self, x: np.ndarray) -> float:
        return float(self.weights.dot(x**2))


def elliptic(x: np.ndarray) -> float:
    return Elliptic(x.size)(x)


def rastrigin(x: np.ndarray) -> float:
    sum_cos = float(np.sum(np.cos(2 * np.pi * x)))
    return 10.0 * (x.size - sum_cos) + sphere(x)


def ackley(x: np.ndarray) -> float:
    dim = x.size
    sum_cos = np.sum(np.cos(2 * np.pi * x))
    return -20.0 * exp(-0.2 * sqrt(sphere(x) / dim)) - exp(sum_cos / dim) + 20 + exp(1)


def schwefel_1_2(x: np.ndarray) -> float:
    cx = np.cumsum(x)
    return sphere(cx)


def rosenbrock(x: np.ndarray) -> float:
    x_m_1 = x[:-1] - 1
    x_diff = x[:-1] ** 2 - x[1:]
    return float(100 * x_diff.dot(x_diff) + x_m_1.dot(x_m_1))


# %% Transformations


def irregularity(x: np.ndarray) -> np.ndarray:  # Tosz
    xhat = np.abs(x)
    xhat[x == 0.0] = 1.0  # get rid of 0 special case
    xhat = np.log(xhat)
    signx = np.sign(x, dtype=float)
    c1 = 0.5 * (10.0 + 5.5) + 0.5 * (10 - 5.5) * signx  # 5.5 if negative else 10
    c2 = 0.5 * (3.1 + 7.9) + 0.5 * (7.9 - 3.1) * signx  # 3.1 if negative else 7.9
    # the following seems slightly faster than positive then negative
    output: np.ndarray = signx * np.exp(xhat + 0.049 * (np.sin(c1 * xhat) + np.sin(c2 * xhat)))
    return output


class Asymmetry:  # Tasy
    def __init__(self, beta: float = 0.2) -> None:
        self.beta = beta
        self._weights: Optional[np.ndarray] = None

    def _get_weights(self, dimension: int) -> np.ndarray:
        if self._weights is None:  # caching for speed
            self._weights = np.linspace(0, self.beta, dimension)
        return self._weights

    def __call__(self, x: np.ndarray) -> np.ndarray:
        exponents = 1 + self._get_weights(x.size) * np.sqrt(np.maximum(0, x))
        return x**exponents  # type: ignore


class Illconditionning:  # Lambda matrix
    def __init__(self, alpha: float = 10.0) -> None:
        self.alpha = alpha
        self._weights: Optional[np.ndarray] = None

    def _get_weights(self, dimension: int) -> np.ndarray:
        if self._weights is None:  # caching for speed
            self._weights = self.alpha ** np.linspace(0, 0.5, num=dimension)
        return self._weights

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self._get_weights(x.size) * x  # type: ignore


class Translation:
    def __init__(self, translation: np.ndarray) -> None:
        self.translation = translation

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return x - self.translation  # type: ignore


class Indexing:
    def __init__(self, indices: np.ndarray, indim: int) -> None:
        self.indices = indices
        self.outdim = indices.size
        self.indim = indim

    def __call__(self, x: np.ndarray) -> np.ndarray:
        if not x.size == self.indim:
            raise ValueError(f"Got size {x.size} but expected {self.indim}")
        return x[self.indices]  # type: ignore

    @classmethod
    def from_split(cls, permutation: np.ndarray, dimensions: List[int], overlap: int = 0) -> List["Indexing"]:
        indexing = split(permutation, dimensions, overlap)
        return [cls(inds, permutation.size) for inds in indexing]


class Rotation:
    def __init__(self, rotation: np.ndarray) -> None:
        shape = rotation.shape
        assert len(shape) == 2 and shape[0] == shape[1]
        self.rotation = rotation

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self.rotation.dot(x)  # type: ignore

    @classmethod
    def from_random(cls, dimension: int, random_state: Optional[np.random.RandomState] = None) -> "Rotation":
        if random_state is None:
            random_state = np.random  # type: ignore
        return cls(np.linalg.qr(random_state.normal(0, 1, size=(dimension, dimension)))[0])  # type: ignore


def split(permutation: np.ndarray, dimensions: List[int], overlap: int = 0) -> List[np.ndarray]:
    expected_size = sum(dimensions) - (len(dimensions) - 1) * overlap
    assert len(permutation) == len(set(permutation))
    assert min(permutation) == 0
    assert max(permutation) == len(permutation) - 1
    if permutation.size != expected_size:
        raise ValueError(
            f"Permutation should have size {expected_size} for dimensions {dimensions} with overlap {overlap}"
        )
    pattern: List[np.ndarray] = []
    start_ind = 0
    for length in dimensions:
        pattern.append(permutation[start_ind : start_ind + length])
        start_ind += length - overlap
    assert start_ind == permutation.size - overlap
    return pattern


def apply_transforms(x: np.ndarray, transforms: List[Callable[[np.ndarray], np.ndarray]]) -> np.ndarray:
    y = transforms[0](x)
    for transf in transforms[1:]:
        y = transf(y)
    return y



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
import pytest
from nevergrad.common import testing
from nevergrad.parametrization import parameter as p
from . import functionlib


DESCRIPTION_KEYS = {
    "split",
    "function_class",
    "name",
    "block_dimension",
    "useful_dimensions",
    "useless_variables",
    "translation_factor",
    "num_blocks",
    "rotation",
    "noise_level",
    "dimension",
    "discrete",
    "aggregator",
    "zero_pen",
    "expo",
    "hashing",
    "parametrization",
    "noise_dissymmetry",
    "bounded",
}


def test_testcase_function_errors() -> None:
    config: tp.Dict[str, tp.Any] = {
        "name": "blublu",
        "block_dimension": 3,
        "useless_variables": 6,
        "num_blocks": 2,
    }
    np.testing.assert_raises(ValueError, functionlib.ArtificialFunction, **config)  # blublu does not exist
    config.update(name="sphere")
    functionlib.ArtificialFunction(**config)  # should wor
    config.update(num_blocks=0)
    np.testing.assert_raises(ValueError, functionlib.ArtificialFunction, **config)  # num blocks should be > 0
    config.update(num_blocks=2.0)
    np.testing.assert_raises(TypeError, functionlib.ArtificialFunction, **config)  # num blocks should be > 0
    config.update(num_blocks=2, rotation=1)
    np.testing.assert_raises(TypeError, functionlib.ArtificialFunction, **config)  # num blocks should be > 0


def test_artitificial_function_repr() -> None:
    config: tp.Dict[str, tp.Any] = {
        "name": "sphere",
        "block_dimension": 3,
        "useless_variables": 6,
        "num_blocks": 2,
    }
    func = functionlib.ArtificialFunction(**config)
    output = repr(func)
    assert "sphere" in output, f"Unexpected representation: {output}"


def test_ptb_no_overfitting() -> None:
    func = functionlib.PBT(("sphere", "cigar"), (3, 7), 12)
    func = func.copy()
    # We do a gradient descent.
    value = [func(-15.0 * np.ones(2)) for _ in range(1500)]
    # We check that the values are becoming better and better.
    assert value[-1] < value[len(value) // 2]  # type: ignore
    assert value[0] > value[len(value) // 2]  # type: ignore


@testing.parametrized(
    sphere=(
        {"name": "sphere", "block_dimension": 3, "useless_variables": 6, "num_blocks": 2},
        13.377591870607294,
    ),
    cigar=(
        {"name": "cigar", "block_dimension": 3, "useless_variables": 6, "num_blocks": 2},
        12492378.626191331,
    ),
    cigar_rot=(
        {"rotation": True, "name": "cigar", "block_dimension": 3, "useless_variables": 6, "num_blocks": 2},
        2575881.272645816,
    ),
    hashed=(
        {"name": "sphere", "block_dimension": 3, "useless_variables": 6, "num_blocks": 2, "hashing": True},
        8.916424986561422,
    ),
    noisy_sphere=(
        {"name": "sphere", "block_dimension": 3, "useless_variables": 6, "num_blocks": 2, "noise_level": 0.2},
        14.512132049518083,
    ),
    noisy_very_sphere=(
        {
            "name": "sphere",
            "block_dimension": 3,
            "useless_variables": 6,
            "num_blocks": 2,
            "noise_dissymmetry": True,
            "noise_level": 0.2,
        },
        19.33566196119778,
    ),
)
def test_testcase_function_value(config: tp.Dict[str, tp.Any], expected: float) -> None:
    # make sure no change is made to the computation
    func = functionlib.ArtificialFunction(**config)
    np.random.seed(1)  # don't know how to control to randomness
    func = func.copy()
    np.random.seed(2)  # initialization is delayed
    x = np.random.normal(0, 1, func.dimension)
    x *= -1 if config.get("noise_dissymmetry", False) else 1  # change sign to activate noise dissymetry
    np.random.seed(12)  # function randomness comes at first call
    value = func(x)
    np.testing.assert_almost_equal(value, expected, decimal=3)


@testing.parametrized(
    random=(np.random.normal(0, 1, 12), False),
    hashed=(np.ones(12), True),
)
def test_test_function(x: tp.Any, hashing: bool) -> None:
    config: tp.Dict[str, tp.Any] = {
        "name": "sphere",
        "block_dimension": 3,
        "useless_variables": 6,
        "num_blocks": 2,
        "hashing": hashing,
    }
    outputs = []
    for _ in range(2):
        np.random.seed(12)
        func = functionlib.ArtificialFunction(**config)
        outputs.append(func(x))
    np.testing.assert_equal(outputs[0], outputs[1])
    # make sure it is properly random otherwise
    outputs.append(functionlib.ArtificialFunction(**config)(x))
    assert outputs[1] != outputs[2]


def test_oracle() -> None:
    func = functionlib.ArtificialFunction("sphere", 5, noise_level=0.1)
    x = np.array([1, 2, 1, 0, 0.5])
    y1 = func(x)  # returns a float
    y2 = func(x)  # returns a different float since the function is noisy
    np.testing.assert_raises(AssertionError, np.testing.assert_array_almost_equal, y1, y2)
    reco = p.Array(init=x)
    y3 = func.evaluation_function(reco)  # returns a float
    # returns the same float (no noise for oracles + sphere function is deterministic)
    y4 = func.evaluation_function(reco)
    np.testing.assert_array_almost_equal(y3, y4)  # should be equal


@pytest.mark.parametrize("split", [True, False])  # type: ignore
def test_blocks(split: bool) -> None:
    func = functionlib.ArtificialFunction("sphere", block_dimension=5, split=split, num_blocks=2)
    assert func.dimension == 10
    param = func.parametrization
    y = func(*param.args, **param.kwargs)
    y2 = func.evaluation_function(param)  # returns a float
    t = func.compute_pseudotime((param.args, param.kwargs), y2)
    assert t > 0
    np.testing.assert_array_almost_equal(y, y2)  # should be equal


def test_function_transform() -> None:
    func = functionlib.ArtificialFunction("sphere", 2, num_blocks=1, noise_level=0.1)
    output = func._transform(np.array([0.0, 0]))
    np.testing.assert_equal(output.shape, (1, 2))
    np.testing.assert_equal(len(output), 1)


def test_artificial_function_summary() -> None:
    func = functionlib.ArtificialFunction("sphere", 5)
    testing.assert_set_equal(func.descriptors.keys(), DESCRIPTION_KEYS)
    np.testing.assert_equal(func.descriptors["function_class"], "ArtificialFunction")


def test_functionlib_copy() -> None:
    func = functionlib.ArtificialFunction("sphere", 5, noise_level=0.2, num_blocks=4)
    func2 = func.copy()
    assert func.equivalent_to(func2)
    assert func._parameters["noise_level"] == func2._parameters["noise_level"]
    assert func is not func2


def test_compute_pseudotime() -> None:
    x = np.array([2.0, 2])
    func = functionlib.ArtificialFunction("sphere", 2)
    np.testing.assert_equal(func.compute_pseudotime(((x,), {}), 3), 1.0)
    np.random.seed(12)
    func = functionlib.ArtificialFunction("DelayedSphere", 2)
    np.testing.assert_almost_equal(func.compute_pseudotime(((x,), {}), 3), 0.00025003021607278633)
    # check minimum
    np.random.seed(None)
    func = functionlib.ArtificialFunction("DelayedSphere", 2)
    func([0, 0])  # trigger init
    x = func.transform_var._transforms[0].translation
    np.testing.assert_equal(func(x), 0)
    np.testing.assert_equal(func.compute_pseudotime(((x,), {}), 0), 0)


@testing.parametrized(
    no_noise=(2, False, False, False),
    noise=(2, True, False, True),
    noise_dissymmetry_pos=(2, True, True, False),  # no noise on right side
    noise_dissymmetry_neg=(-2, True, True, True),
    no_noise_with_dissymmetry_neg=(-2, False, True, False),
)
def test_noisy_call(x: int, noise: bool, noise_dissymmetry: bool, expect_noisy: bool) -> None:
    fx = functionlib._noisy_call(
        x=np.array([x]),
        transf=np.tanh,
        func=lambda y: np.arctanh(y)[0],  # type: ignore
        noise_level=float(noise),
        noise_dissymmetry=noise_dissymmetry,
        random_state=np.random.RandomState(0),
    )
    assert not np.isnan(fx)  # noise addition should not get out of function domain
    if expect_noisy:
        np.testing.assert_raises(AssertionError, np.testing.assert_almost_equal, fx, x, decimal=8)
    else:
        np.testing.assert_almost_equal(fx, x, decimal=8)


@pytest.mark.parametrize("independent_sigma", [True, False])  # type: ignore
@pytest.mark.parametrize("mutable_sigma", [True, False])  # type: ignore
def test_far_optimum_function(independent_sigma: bool, mutable_sigma: bool) -> None:
    func = functionlib.FarOptimumFunction(
        independent_sigma=independent_sigma, mutable_sigma=mutable_sigma
    ).copy()
    param = func.parametrization.spawn_child()
    assert isinstance(param, p.Array)
    assert isinstance(param.sigma, p.Array) == mutable_sigma
    assert param.sigma.value.size == (1 + independent_sigma)  # type: ignore
    param.mutate()
    new_val = param.sigma.value
    assert bool(np.sum(np.abs(new_val - func.parametrization.sigma.value))) == mutable_sigma  # type: ignore
    if independent_sigma and mutable_sigma:
        assert new_val[0] != new_val[1]  # type: ignore


def test_far_optimum_function_cases() -> None:
    cases = list(functionlib.FarOptimumFunction.itercases())
    assert len(cases) == 48



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import numpy as np
import nevergrad as ng
from . import core


def test_causal_discovery_using_data_func() -> None:
    np.random.seed(12)
    func = core.CausalDiscovery(generator="sachs")
    assert func._nvars == 11
    # Prepare recommendation
    param_links = ng.p.Choice([-1, 0, 1], repetitions=func._nvars * (func._nvars - 1) // 2)
    param_links.set_standardized_data([i % 2 for i in range(0, param_links.dimension)])
    result = func(network_links=param_links.value)
    assert np.isclose(36.721185206294926, result, atol=1e-10)


def test_causal_discovery_using_data_minimize() -> None:
    # Optimization should return the same result since the true graph is not random and small
    np.random.seed(12)
    func = core.CausalDiscovery(generator="sachs")
    optimizer = ng.optimizers.OnePlusOne(parametrization=func.parametrization, budget=500)
    optimizer.parametrization.random_state = np.random.RandomState(12)
    recommendation = optimizer.minimize(func)
    recommendation_score = func(**recommendation.kwargs)
    assert np.isclose(20.498524203069657, recommendation_score, atol=1e-10)  # Optimal
    assert len(recommendation.kwargs["network_links"]) == func._nvars * (func._nvars - 1) // 2


def test_causal_discovery_using_generator() -> None:
    nnodes = 13
    npoints = 55
    func = core.CausalDiscovery(generator="acylicgraph", npoints=npoints, nodes=nnodes)
    assert func._nvars == nnodes
    assert func._data.shape == (npoints, nnodes)
    assert func.graph_score(func._ground_truth_graph) == 1



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Nevergrad's API provides ask/tell, letting the user run the objective function
in their own way. Some libraries for optimization take the objective function
as a parameter and run it themselves.
This module is a system for adapting such a library to be used in nevergrad,
by running the optimization on a background thread.
The file recastlib.py has some uses of it.
For instructions for doing this for another library, see
SequentialRecastOptimizer.
"""


import warnings
import threading
import queue
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common.errors import NevergradError
from nevergrad.parametrization import parameter as p
from . import base
from .base import IntOrParameter


class StopOptimizerThread(Exception):
    pass


class TooManyAskError(NevergradError):
    pass


class _MessagingThread(threading.Thread):
    """Thread that runs a function taking another function as input. Each call of the inner function
    adds the point given by the algorithm into the ask queue and then blocks until the main thread sends
    the result back into the tell queue.

    Note
    ----
    This thread must be overlaid into another MessagingThread  because:
    - the threading part should hold no reference from outside (otherwise the destructors may wait each other)
    - the destructor cannot be implemented, hence there is no way to stop the thread automatically
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, caller: tp.Callable[..., tp.Any], *args: tp.Any, **kwargs: tp.Any) -> None:
        super().__init__()
        self.messages_ask: tp.Any = queue.Queue()
        self.messages_tell: tp.Any = queue.Queue()
        self.call_count = 0
        self.error: tp.Optional[Exception] = None
        self._caller = caller
        self._args = args
        self._kwargs = kwargs
        self.output: tp.Optional[tp.Any] = None  # TODO add a "done" attribute ?

    def run(self) -> None:
        """Starts the thread and run the "caller" function argument on
        the fake callable, which posts messages and awaits for their answers.
        """
        try:
            self.output = self._caller(self._fake_callable, *self._args, **self._kwargs)
        except StopOptimizerThread:  # gracefully stopping the thread
            self.messages_ask.put(ValueError("Optimization told to finish"))
        except Exception as e:  # pylint: disable=broad-except
            self.messages_ask.put(e)
            self.error = e
        else:
            self.messages_ask.put(None)

    def _fake_callable(self, *args: tp.Any) -> tp.Any:
        """
        Puts a new point into the ask queue to be evaluated on the
        main thread and blocks on get from tell queue until point
        is evaluated on main thread and placed into tell queue when
        it is then returned to the caller.
        """
        self.call_count += 1
        self.messages_ask.put(args[0])  # sends a message
        candidate = self.messages_tell.get()  # get evaluated message
        if candidate is None:
            raise StopOptimizerThread()
        return candidate

    def stop(self) -> None:
        """Notifies the thread that it must stop"""
        self.messages_tell.put(None)


class MessagingThread:
    """Encapsulate the inner thread, so that kill order is automatically called at deletion."""

    def __init__(self, caller: tp.Callable[..., tp.Any], *args: tp.Any, **kwargs: tp.Any) -> None:
        self._thread = _MessagingThread(caller, *args, **kwargs)
        self._thread.start()

    def is_alive(self) -> bool:
        return self._thread.is_alive()

    @property
    def output(self) -> tp.Any:
        return self._thread.output

    @property
    def error(self) -> tp.Optional[Exception]:
        return self._thread.error

    @property
    def messages_tell(self) -> tp.Any:
        return self._thread.messages_tell

    @property
    def messages_ask(self) -> tp.Any:
        return self._thread.messages_ask

    def stop(self) -> None:
        self._thread.stop()

    def __del__(self) -> None:
        self.stop()  # del method of the thread class does not work


class RecastOptimizer(base.Optimizer):
    """Base class for ask and tell optimizer derived from implementations with no ask and tell interface.
    The underlying optimizer implementation is a function which is supposed to call directly the function
    to optimize. It is tricked into optimizing a "fake" function in a thread:
    - calls to the fake functions are returned by the "ask()" interface
    - return values of the fake functions are provided to the thread when calling "tell(x, value)"

    Note
    ----
    These implementations are not necessarily robust. More specifically, one cannot "tell" any
    point which was not "asked" before.

    An optimization is performed by a third-party library in a background thread. This communicates
    with the main thread using two queue objects. Specifically:

        messages_ask is filled by the background thread with a candidate (or batch of candidates)
        it wants evaluated for, or None if the background thread is over, or an Exception
        which needs to be raised to the user.

        messages_tell supplies the background thread with a value to return from the fake function.
        A value of None means the background thread is no longer relevant and should exit.
    """

    recast = True

    def __init__(
        self,
        parametrization: IntOrParameter,
        budget: tp.Optional[int] = None,
        num_workers: int = 1,
    ) -> None:
        super().__init__(parametrization, budget, num_workers=num_workers)
        self._messaging_thread: tp.Optional[MessagingThread] = None  # instantiate at runtime

    def get_optimization_function(self) -> tp.Callable[[tp.Callable[..., tp.Any]], tp.Optional[tp.ArrayLike]]:
        """Return an optimization procedure function (taking a function to optimize as input)

        Note
        ----
        This optimization procedure must be a function or an object which is completely
        independent from self, otherwise deletion of the optimizer may hang indefinitely.
        """
        raise NotImplementedError(
            "You should define your optimizer! Also, be very careful to avoid "
            " reference to this instance in the returned object"
        )

    def _check_error(self) -> None:
        if self._messaging_thread is not None:
            if self._messaging_thread.error is not None:
                raise RuntimeError(
                    f"Recast optimizer raised an error:\n{self._messaging_thread.error}"
                ) from self._messaging_thread.error

    def _post_loss(self, candidate: p.Parameter, loss: float) -> tp.Loss:
        # pylint: disable=unused-argument
        """
        Posts the value, and the thread will deal with it.
        """
        return loss

    def _internal_tell_not_asked(self, candidate: p.Parameter, loss: float) -> None:
        raise base.errors.TellNotAskedNotSupportedError

    def _internal_provide_recommendation(self) -> tp.Optional[tp.ArrayLike]:
        """Returns the underlying optimizer output if provided (ie if the optimizer did finish)
        else the best pessimistic point.
        """
        if self._messaging_thread is not None and self._messaging_thread.output is not None:
            return self._messaging_thread.output
        else:
            return None  # use default

    def __del__(self) -> None:
        # explicitly ask the thread to stop (better be safe :))
        if self._messaging_thread is not None:
            self._messaging_thread.stop()


class SequentialRecastOptimizer(RecastOptimizer):
    """Recast Optimizer which cannot deal with parallelization

    There can only be one worker. Each ask must be followed by
    a tell.

    A simple usage is that you have a library which can minimize
    a function which returns a scalar.
    Just make an optimizer inheriting from this class, and inplement
    get_optimization_function to return a callable which runs the
    optimization, taking the objective as its only parameter. The
    callable must not have any references to the optimizer itself.
    (This avoids a reference cycle between the background thread and
    the optimizer, aiding cleanup.) It can have a weakref though.

    If you want your optimizer instance to be picklable, we have to
    store every candidate during optimization, which may use a lot
    of memory. This lets us replay the optimization when
    unpickling. We only do this if you ask for it. To enable:
        - The optimization must be reproducible, asking for the same
          candidates every time. If you need a seed from nevergrad's
          generator, you can't necessarily generate this again after
          unpickling. One solution is to store it in self, if it is
          not there yet, in the body of get_optimization_function.

          As in general in nevergrad, do not set the seed from the
          RNG in your own __init__ because it will cause surprises
          to anyone re-seeding your parametrization after init.
        - The user must call enable_pickling() after initializing
          the optimizer instance.
    """

    # pylint: disable=abstract-method

    no_parallelization = True

    def __init__(
        self,
        parametrization: IntOrParameter,
        budget: tp.Optional[int],
        num_workers: int = 1,
    ) -> None:
        super().__init__(parametrization=parametrization, budget=budget, num_workers=num_workers)
        self._enable_pickling = False
        self.replay_archive_tell: tp.List[p.Parameter] = []

    def enable_pickling(self):
        """Make the optimizer store its history of tells, so
        that it can be serialized.
        """
        if self.num_ask != 0:
            raise ValueError("Can only enable pickling before all asks.")
        self._enable_pickling = True

    def _internal_ask_candidate(self) -> p.Parameter:
        """Reads messages from the thread in which the underlying optimization function is running
        New messages are sent as "ask".
        """
        if self._messaging_thread is None:
            self._messaging_thread = MessagingThread(self.get_optimization_function())
        alive = self._messaging_thread.is_alive()
        if alive:
            point = self._messaging_thread.messages_ask.get()
            if isinstance(point, Exception):
                raise point
        if not alive or point is None:  # In case the algorithm stops before the budget is elapsed.
            warnings.warn(
                "Underlying optimizer has already converged, returning random points",
                base.errors.FinishedUnderlyingOptimizerWarning,
            )
            self._check_error()
            data = self._rng.normal(0, 1, self.dimension)
            return self.parametrization.spawn_child().set_standardized_data(data)
        candidate = self.parametrization.spawn_child().set_standardized_data(point)
        return candidate

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: float) -> None:
        """Returns value for a point which was "asked"
        (none asked point cannot be "tell")
        """
        assert self._messaging_thread is not None, 'Start by using "ask" method, instead of "tell" method'
        if not self._messaging_thread.is_alive():  # optimizer is done
            self._check_error()
            return
        if self._enable_pickling:
            self.replay_archive_tell.append(candidate)
        self._messaging_thread.messages_tell.put(self._post_loss(candidate, loss))

    def __getstate__(self):
        if not self._enable_pickling:
            raise ValueError("If you want picklability you should have asked for it")
        thread = self._messaging_thread
        self._messaging_thread = None
        state = self.__dict__.copy()
        self._messaging_thread = thread
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if not self._enable_pickling:
            raise ValueError("Cannot unpickle the unpicklable")

        # We temporarily unset _enable_pickling so that the replays do not
        # get archived again.
        self._enable_pickling = False
        for i, candidate in enumerate(self.replay_archive_tell):
            new_candidate = self._internal_ask_candidate()
            norm = np.linalg.norm(new_candidate.get_standardized_data(reference=candidate))
            # Check that the replay wants the same value as we had the first time.
            # If an error is raised here then you might want to
            # check the reproducibility of your optimizer.
            if norm > 0.00001:
                raise RuntimeError(f"Mismatch in replay at index {i} of {len(self.replay_archive_tell)}.")
            self._internal_tell_candidate(candidate, candidate.loss)

        if self.num_ask > self.num_tell:
            self._internal_ask_candidate()

        self._enable_pickling = True


class BatchRecastOptimizer(RecastOptimizer):
    """Recast optimizer where points to evaluate are provided in batches
    and stored by the optimizer to be asked and told on. The fake_callable
    is only brought into action every 'batch size' number of asks and tells
    instead of every ask and tell. This opens up the optimizer to
    parallelism.

    Note
    ----
    You have to complete a batch before you start a new one so parallelism
    is only possible within batches i.e. if a batch size is 100 and you have
    done 100 asks, you must do 100 tells before you ask again but you could do
    those 100 asks and tells in parallel. To find out if you can perform an ask
    at any given time call self.can_ask.
    """

    # pylint: disable=abstract-method

    def __init__(
        self, parametrization: IntOrParameter, budget: tp.Optional[int] = None, num_workers: int = 1
    ) -> None:
        super().__init__(parametrization, budget, num_workers=num_workers)
        self._current_batch: tp.List[p.Parameter] = []
        self._batch_losses: tp.List[tp.Loss] = []
        self._tell_counter = 0
        self.batch_size = 0
        self.indices: tp.Dict[str, int] = {}

    def _internal_ask_candidate(self) -> p.Parameter:
        """Reads messages from the thread in which the underlying optimization function is running
        New messages are sent as "ask".
        """
        if self._messaging_thread is None:
            self._messaging_thread = MessagingThread(self.get_optimization_function())
        if self._current_batch:
            return self._current_batch.pop()
        # if there are any points in the previous batch that haven't been told on, you cannot update the current batch.
        if not self.can_ask():
            raise TooManyAskError(
                "You can't get a new batch until the old one has been fully told on. See docstring for more info."
            )
        alive = self._messaging_thread.is_alive()
        # wait for a message
        if alive:
            points = self._messaging_thread.messages_ask.get()
            if isinstance(points, Exception):
                raise points
        if not alive or points is None:  # In case the algorithm stops before the budget is elapsed.
            warnings.warn(
                "Underlying optimizer has already converged, returning random points",
                base.errors.FinishedUnderlyingOptimizerWarning,
            )
            self._check_error()
            data = self._rng.normal(0, 1, self.dimension)
            return self.parametrization.spawn_child().set_standardized_data(data)

        # We have a new batch
        self.batch_size = len(points)
        self._current_batch = [
            self.parametrization.spawn_child().set_standardized_data(point) for point in points
        ]
        self._batch_losses = [None] * len(points)  # type: ignore
        # map each point to an index in preparation to build loss array in tell
        self.indices = {candidate.uid: i for i, candidate in enumerate(self._current_batch)}
        return self._current_batch.pop()

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: float) -> None:
        """Returns value for a point which was "asked"
        (none asked point cannot be "tell")
        """
        assert self._messaging_thread is not None, 'Start by using "ask" method, instead of "tell" method'
        if not self._messaging_thread.is_alive():  # optimizer is done
            self._check_error()
            return
        candidate_index = self.indices.pop(candidate.uid)
        self._batch_losses[candidate_index] = self._post_loss(candidate, loss)
        self._tell_counter += 1
        # if batch size number of tells since new batch, send array of losses to fake callable
        if self._tell_counter == self.batch_size:
            self._messaging_thread.messages_tell.put(np.array(self._batch_losses))
            self._batch_losses = []
            self._tell_counter = 0

    def minimize(
        self,
        objective_function: tp.Callable[..., tp.Loss],
        executor: tp.Optional[tp.ExecutorLike] = None,
        batch_mode: bool = False,
        verbosity: int = 0,
        constraint_violation: tp.Any = None,
    ) -> p.Parameter:
        raise NotImplementedError("This optimizer isn't supported by the way minimize works by default.")

    def can_ask(self) -> bool:
        """Returns whether the optimizer is able to perform another ask,
        either because there are points left in the current batch to ask
        or you are ready for a new batch (You have asked and told on every
        point in the last batch.)
        """
        return len(self.indices) == 0 or len(self._current_batch) > 0



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
from unittest import SkipTest
import numpy as np
import pytest
import nevergrad.common.typing as tp
import nevergrad as ng
from .. import base
from ..optimizerlib import registry
from . import core


def test_hypervolume_pareto_function() -> None:
    hvol = core.HypervolumePareto(upper_bounds=(100, 100))
    tuples = [
        (110, 110),  # -0 + distance
        (110, 90),  # -0 + distance
        (80, 80),  # -400 + distance
        (50, 50),  # -2500 + distance
        (50, 50),  # -2500 + distance
        (80, 80),  # -2500 + distance --> -2470
        (30, 60),  # [30,50]x[60,100] + [50,100]x[50,100] --> -2500 -800 = -3300
        (60, 30),
    ]  # [30,50]x[60,100] + [50,100]x[50,100] + [60,100]x[30,50] --> -2500 -800 -800= -4100
    values = []
    for tup in tuples:
        param = ng.p.Tuple(*(ng.p.Scalar(x) for x in tup))
        param._losses = np.array(tup)
        values.append(hvol.add(param))
    expected = [20, 10, -400, -2500.0, -2500.0, -2470.0, -3300.0, -4100.0]
    assert values == expected, f"Expected {expected} but got {values}"
    front = [p.value for p in hvol.pareto_front()]
    expected_front = [(50, 50), (30, 60), (60, 30)]
    assert front == expected_front, f"Expected {expected_front} but got {front}"


def test_hypervolume_pareto_with_no_good_point() -> None:
    hvol = core.HypervolumePareto(upper_bounds=(100, 100))
    tuples = [(110, 110), (110, 90), (90, 110)]
    values = []
    for tup in tuples:
        param = ng.p.Tuple(*(ng.p.Scalar(x) for x in tup))
        param._losses = np.array(tup)
        values.append(hvol.add(param))
    expected = [20, 10, 10]
    assert values == expected, f"Expected {expected} but got {values}"
    front = [p.value for p in hvol.pareto_front()]
    expected_front = [(110, 90), (90, 110)]
    assert front == expected_front, f"Expected {expected_front} but got {front}"
    assert hvol._best_volume == -10  # used in xps


@pytest.mark.parametrize("losses", [(12, [3, 4]), ([3, 4], 12)])  # type: ignore
def test_num_losses_error(losses: tp.Tuple[tp.Any, tp.Any]) -> None:
    opt = ng.optimizers.CMA(parametrization=3, budget=100)
    cand = opt.ask()
    opt.tell(cand, losses[0])
    cand = opt.ask()
    with pytest.raises(ValueError):
        opt.tell(cand, losses[1])


def mofunc(array: np.ndarray) -> np.ndarray:
    return abs(array - 1)  # type: ignore


@pytest.mark.parametrize("name", registry)  # type: ignore
def test_optimizers_multiobjective(name: str) -> None:  # pylint: disable=redefined-outer-name
    def doint(s):  # Converting a string into an int.
        return 7 + sum([ord(c) * i for i, c in enumerate(s)])

    if doint(name) % 5 > 0 or "BO" in name or "AX" in name or "pysot" in name:
        raise SkipTest("BO is currently failing for unclear reasons")  # TODO solve
    with warnings.catch_warnings():
        # tests do not need to be efficient
        warnings.simplefilter("ignore", category=base.errors.InefficientSettingsWarning)
        optimizer = registry[name](parametrization=4, budget=100)
        candidate = optimizer.ask()
        optimizer.tell(candidate, mofunc(candidate.value))
    # to be tested
    optimizer.ask()
    to_be_tested = {"DE"}  # specific mo adapatation
    assert not to_be_tested - set(registry), "Testing some unknown optimizers"
    if optimizer.name in to_be_tested:
        optimizer.minimize(mofunc)


# pylint: disable=redefined-outer-name,unsubscriptable-object,
# pylint: disable=unused-variable,unused-import,reimported,import-outside-toplevel
def test_doc_multiobjective() -> None:
    # DOC_MULTIOBJ_OPT_0
    import nevergrad as ng
    import numpy as np

    def multiobjective(x):
        return [np.sum(x**2), np.sum((x - 1) ** 2)]

    print("Example: ", multiobjective(np.array([1.0, 2.0, 0])))
    # >>> Example: [5.0, 2.0]

    optimizer = ng.optimizers.CMA(parametrization=3, budget=100)

    # for all but DE optimizers, deriving a volume out of the losses,
    # it's not strictly necessary but highly advised to provide an
    # upper bound reference for the losses (if not provided, such upper
    # bound is automatically inferred with the first few "tell")
    optimizer.tell(ng.p.MultiobjectiveReference(), [5, 5])
    # note that you can provide a Parameter to MultiobjectiveReference,
    # which will be passed to the optimizer

    optimizer.minimize(multiobjective, verbosity=2)

    # The function embeds its Pareto-front:
    print("Pareto front:")
    for param in sorted(optimizer.pareto_front(), key=lambda p: p.losses[0]):
        print(f"{param} with losses {param.losses}")

    # >>> Array{(3,)}:[0. 0. 0.] with loss [0. 3.]
    #     Array{(3,)}:[0.39480968 0.98105712 0.55785803] with loss [1.42955333 0.56210368]
    #     Array{(3,)}:[1.09901515 0.97673712 0.97153943] with loss [3.10573857 0.01115516]

    # It can also provide subsets:
    print("Random subset:", optimizer.pareto_front(2, subset="random"))
    print("Loss-covering subset:", optimizer.pareto_front(2, subset="loss-covering"))
    print("Domain-covering subset:", optimizer.pareto_front(2, subset="domain-covering"))
    print("EPS subset:", optimizer.pareto_front(2, subset="EPS"))

    # DOC_MULTIOBJ_OPT_1
    assert len(optimizer.pareto_front()) > 1
    assert len(optimizer.pareto_front(2, "loss-covering")) == 2
    assert len(optimizer.pareto_front(2, "domain-covering")) == 2
    assert len(optimizer.pareto_front(2, "hypervolume")) == 2
    assert len(optimizer.pareto_front(2, "random")) == 2
    assert len(optimizer.pareto_front(2, "EPS")) == 2



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
from pathlib import Path
from concurrent import futures
import nevergrad.common.typing as tp
from . import utils
from . import core
from . import plotting


# pylint: disable=too-many-arguments
def launch(
    experiment: str,
    num_workers: int = 1,
    seed: tp.Optional[int] = None,
    cap_index: tp.Optional[int] = None,
    output: tp.Optional[tp.PathLike] = None,
) -> Path:
    """Launch experiment with given names and selection modulo
    cap_index can be specified to provide a limited number of settings
    """
    # create the data
    csvpath = Path(experiment + ".csv") if output is None else Path(output)
    if num_workers == 1:
        df = core.compute(experiment, cap_index=cap_index, seed=seed)
    else:
        with futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
            df = core.compute(
                experiment, seed=seed, cap_index=cap_index, executor=executor, num_workers=num_workers
            )
    # save data to csv
    try:
        core.save_or_append_to_csv(df, csvpath)
    except Exception:  # pylint: disable=broad-except
        csvpath = Path(experiment + ".csv")
        print(f"Failed to save to {output}, falling back to {csvpath}")
        core.save_or_append_to_csv(df, csvpath)
    else:
        print(f"Saved data to {csvpath}")
    return csvpath


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Run an experiment and create a result csv file.")
    parser.add_argument(
        "experiment", type=str, help="name of an experiment registered in the experiments registry"
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=None,
        help="Use a seed for reproducibility (for generators which take care of seeding)",
    )
    parser.add_argument(
        "--cap_index", type=int, default=None, help="Stop after generationg/running settings #cap_index"
    )
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="Output path for the CSV file (default: <experiment>.csv). Existing files are appended",
    )
    parser.add_argument(
        "--imports",
        type=str,
        default=None,
        help="Comma-separated list of file paths with additional experiment(s) and/or optimizer(s) definitions",
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        default=1,
        help="Numbers of workers to use for the computation (splits the job in chunks)",
    )
    parser.add_argument(
        "--repetitions",
        type=int,
        default=1,
        help="Number of repetitions to perform for the experiment plan (seeds will be incremented)",
    )
    parser.add_argument(
        "--plot",
        nargs="?",
        default=False,
        const=True,
        help="Creates the corresponding plots if present (provide a path, or folder <experiment>_plots will be used)",
    )
    return parser.parse_args()


def repeated_launch(
    experiment: str,
    num_workers: int = 1,
    seed: tp.Optional[int] = None,
    cap_index: tp.Optional[int] = None,
    output: tp.Optional[tp.PathLike] = None,
    plot: tp.Union[bool, tp.PathLike] = False,
    imports: tp.Optional[tp.List[tp.PathLike]] = None,
    repetitions: int = 1,
) -> None:
    """Launch experiment with given names and selection module
    cap_index can be specified to provide a limited number of settings
    This repeats the plan several times and increments the seed.
    """
    # start by importing additional content
    if imports is not None:
        assert isinstance(imports, (tuple, list))
        for path in imports:
            core.import_additional_module(path)
    # then run multiple times
    csvpath = Path("default.csv")
    for k in range(repetitions):
        print(f"Starting repetition {k +1} / {repetitions}")
        csvpath = launch(
            experiment,
            num_workers=num_workers,
            cap_index=cap_index,
            output=output,
            seed=None if seed is None else seed + k,
        )
    # save plots if need be
    if plot:
        df = utils.Selector.read_csv(csvpath)
        if isinstance(plot, bool):
            plot = str(Path(csvpath).with_suffix("")) + "_plots"
        print(f"Saving plots into folder: {plot}")
        plotting.create_plots(df, output_folder=plot)


if __name__ == "__main__":
    args = get_args()
    repeated_launch(
        args.experiment,
        num_workers=args.num_workers,
        cap_index=args.cap_index,
        output=args.output,
        seed=args.seed,
        plot=args.plot,
        imports=args.imports if args.imports is None else args.imports.split(","),
        repetitions=args.repetitions,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .teampursuit import teampursuit
import math


class Cyclist:

    # static class variables
    max_power = 1200
    min_power = 200
    drag_coeffiecient = 0.65
    mechanical_efficiency = 0.977
    bike_mass = 7.7
    fatigue_level = 0

    def __init__(self, height, weight, mean_maximum_power, event, start_position, gender):

        self.height = height
        self.weight = weight
        self.mean_maximum_power = mean_maximum_power
        self.event = event
        self.start_position = start_position
        self.position = start_position
        self.gender = gender
        self.current_velocity = 0.0
        self.fatigue_level = 0

        self.update_cda()
        self.update_total_energy()
        self.remaining_energy = self.total_energy

    def set_pace(self, power):
        fatigue_factor = 1 - (0.01 * self.fatigue_level)

        delta_ke = (
            (power * self.mechanical_efficiency * fatigue_factor)
            - (self.coefficient_drag_area * 0.5 * self.event.air_density * math.pow(self.current_velocity, 3))
            - (
                teampursuit.friction_coefficient
                * (self.weight + self.bike_mass)
                * teampursuit.gravitational_acceleration
                * self.current_velocity
            )
        ) * teampursuit.time_step

        new_velocity = math.pow(
            ((2 * delta_ke / (self.weight + self.bike_mass)) + math.pow(self.current_velocity, 2)), 0.5
        )
        acceleration = new_velocity - self.current_velocity
        distance = (self.current_velocity * teampursuit.time_step) + (
            0.5 * acceleration * math.pow(teampursuit.time_step, 2)
        )

        self.current_velocity = new_velocity

        if self.remaining_energy > power * teampursuit.time_step:
            self.remaining_energy -= power * teampursuit.time_step
        else:
            self.remaining_energy = 0.0

        return distance

    def follow(self, distance):
        fatigue_factor = 1 - (0.01 * self.fatigue_level)

        acceleration = (
            2
            * (distance - (self.current_velocity * teampursuit.time_step))
            / math.pow(teampursuit.time_step, 2)
        )
        new_velocity = self.current_velocity + (acceleration * teampursuit.time_step)
        delta_ke = 0.5 * (self.weight + self.bike_mass) * (new_velocity - self.current_velocity)
        power = (
            (
                self.coefficient_drag_area
                * teampursuit.drafting_coefficients[self.position - 2]
                * 0.5
                * self.event.air_density
                * math.pow(self.current_velocity, 3)
            )
            + (
                teampursuit.friction_coefficient
                * (self.weight + self.bike_mass)
                * teampursuit.gravitational_acceleration
                * self.current_velocity
            )
            + (delta_ke / teampursuit.time_step)
        ) / (self.mechanical_efficiency * fatigue_factor)

        self.current_velocity = new_velocity

        if self.remaining_energy > power * teampursuit.time_step:
            self.remaining_energy -= power * teampursuit.time_step
        else:
            self.remaining_energy = 0.0

    def get_height(self):
        return self.height

    def get_weight(self):
        return self.weight

    def get_mean_maximum_power(self):
        return self.mean_maximum_power

    def get_remaining_energy(self):
        return self.remaining_energy

    def get_position(self):
        return self.position

    def set_weight(self, weight):
        self.weight = weight
        self.update_cda()
        self.update_total_energy()

    def set_height(self, height):
        self.height = height
        self.update_cda()

    def set_mean_maximum_power(self, mean_maximum_power):
        self.mean_maximum_power = mean_maximum_power
        self.update_total_energy()

    def set_position(self, position):
        self.position = position

    def increase_fatigue(self):
        self.fatigue_level += 2

    def recover(self):
        if self.fatigue_level > 0:
            self.fatigue_level -= 1

    def reset(self):
        self.remaining_energy = self.total_energy
        self.position = self.start_position
        self.fatigue_level = 0
        self.current_velocity = 0

    def update_cda(self):
        self.coefficient_drag_area = self.drag_coeffiecient * (
            (0.0293 * math.pow(self.height, 0.725)) * (math.pow(self.weight, 0.425)) + 0.0604
        )

    def update_total_energy(self):
        coeff = 240 if self.gender == "male" else 210

        self.total_energy = self.mean_maximum_power * self.weight * coeff



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p
from . import metamodel
from . import base
from . import oneshot


class Crossover:
    def __init__(
        self,
        random_state: np.random.RandomState,
        crossover: tp.Union[str, float],
        parameter: tp.Optional[p.Parameter] = None,
    ):
        self.CR = 0.5
        self.crossover = crossover
        self.random_state = random_state
        if isinstance(crossover, float):
            self.CR = crossover
        elif crossover == "random":
            self.CR = self.random_state.uniform(0.0, 1.0)
        elif crossover not in ["twopoints", "onepoint", "rotated_twopoints", "voronoi"]:
            raise ValueError(f'Unknown crossover "{crossover}"')
        self.shape = (
            np.array(parameter.value).shape if (crossover == "voronoi" and parameter is not None) else None
        )

    def apply(self, donor: np.ndarray, individual: np.ndarray) -> None:
        dim = donor.size
        if self.crossover == "twopoints" and dim >= 4:
            return self.twopoints(donor, individual)
        elif self.crossover == "rotated_twopoints" and dim >= 4:
            return self.rotated_twopoints(donor, individual)
        elif self.crossover == "onepoint" and dim >= 3:
            return self.onepoint(donor, individual)
        elif self.crossover == "voronoi":
            return self.voronoi(donor, individual)
        else:
            return self.variablewise(donor, individual)

    def variablewise(self, donor: np.ndarray, individual: np.ndarray) -> None:
        R = self.random_state.randint(donor.size)
        # the following could be updated to vectorial uniform sampling (changes recomms)
        transfer = np.array(
            [idx != R and self.random_state.uniform(0, 1) > self.CR for idx in range(donor.size)]
        )
        donor[transfer] = individual[transfer]

    def onepoint(self, donor: np.ndarray, individual: np.ndarray) -> None:
        R = self.random_state.randint(1, donor.size)
        if self.random_state.choice([True, False]):
            donor[R:] = individual[R:]
        else:
            donor[:R] = individual[:R]

    def twopoints(self, donor: np.ndarray, individual: np.ndarray) -> None:
        bounds = sorted(self.random_state.choice(donor.size + 1, size=2, replace=False).tolist())
        if bounds[1] == donor.size and not bounds[0]:  # make sure there is at least one point crossover
            bounds[self.random_state.randint(2)] = self.random_state.randint(1, donor.size)
        if self.random_state.choice([True, False]):
            donor[bounds[0] : bounds[1]] = individual[bounds[0] : bounds[1]]
        else:
            donor[: bounds[0]] = individual[: bounds[0]]
            donor[bounds[1] :] = individual[bounds[1] :]

    def rotated_twopoints(self, donor: np.ndarray, individual: np.ndarray) -> None:
        bounds = sorted(self.random_state.choice(donor.size + 1, size=2, replace=False).tolist())
        if bounds[1] == donor.size and not bounds[0]:  # make sure there is at least one point crossover
            bounds[self.random_state.randint(2)] = self.random_state.randint(1, donor.size)
        bounds2 = [self.random_state.choice(donor.size + 1 - bounds[1] + bounds[0])]
        bounds2.append(bounds2[0] + bounds[1] - bounds[0])
        assert bounds[1] < donor.size + 1
        donor[bounds[0] : bounds[1]] = individual[bounds2[0] : bounds2[1]]

    def voronoi(self, donor: np.ndarray, individual: np.ndarray) -> None:
        shape = self.shape
        if shape is None or len(shape) < 2:
            warnings.warn("Voronoi DE needs a shape.")
            self.twopoints(donor, individual)
            return
        local_donor = donor.reshape(shape)
        local_individual = individual.reshape(shape)
        x1 = np.array([np.random.randint(shape[i]) for i in range(len(shape))])
        x2 = np.array([np.random.randint(shape[i]) for i in range(len(shape))])
        x3 = np.array([np.random.randint(shape[i]) for i in range(len(shape))])
        x4 = np.array([np.random.randint(shape[i]) for i in range(len(shape))])
        it = np.nditer(local_donor, flags=["multi_index"])
        for _ in it:
            d1 = np.linalg.norm(np.array(it.multi_index) - x1)
            d2 = np.linalg.norm(np.array(it.multi_index) - x2)
            d3 = np.linalg.norm(np.array(it.multi_index) - x3)
            d4 = np.linalg.norm(np.array(it.multi_index) - x4)
            if min([d1, d2, d3]) > d4:
                local_donor[it.multi_index] = local_individual[it.multi_index]
        donor[:] = local_donor.flatten()[:]
        individual[:] = local_individual.flatten()[:]


class _DE(base.Optimizer):
    """Differential evolution.

    Default pop size equal to 30
    We return the mean of the individuals with fitness better than median, which might be stupid sometimes.
    CR =.5, F1=.8, F2=.8, curr-to-best.
    Initial population: pure random.
    """

    # pylint: disable=too-many-locals, too-many-nested-blocks,too-many-instance-attributes
    # pylint: disable=too-many-branches, too-many-statements, too-many-arguments

    def __init__(
        self,
        parametrization: base.IntOrParameter,
        budget: tp.Optional[int] = None,
        num_workers: int = 1,
        config: tp.Optional["DifferentialEvolution"] = None,
        weights: tp.Any = None,
    ) -> None:
        super().__init__(parametrization, budget=budget, num_workers=num_workers)
        # config
        self.objective_weights = weights
        self._config = DifferentialEvolution() if config is None else config
        self.scale = (
            float(1.0 / np.sqrt(self.dimension))
            if isinstance(self._config.scale, str)
            else self._config.scale
        )
        pop_choice = {
            "standard": 0,
            "small": 1 + int(np.sqrt(np.log(self.dimension + 3))),
            "dimension": self.dimension + 1,
            "large": 7 * self.dimension,
        }
        if isinstance(self._config.popsize, int):
            self.llambda = self._config.popsize
        else:
            self.llambda = max(30, self.num_workers, pop_choice[self._config.popsize])
        # internals
        # if budget is not None and budget < 60:
        #    warnings.warn(
        #        "DE algorithms are inefficient with budget < 60", base.errors.InefficientSettingsWarning
        #    )
        self._MULTIOBJECTIVE_AUTO_BOUND = max(self._MULTIOBJECTIVE_AUTO_BOUND, self.llambda)  # type: ignore
        self._penalize_cheap_violations = True
        self._uid_queue = base.utils.UidQueue()
        self.population: tp.Dict[str, p.Parameter] = {}
        self.sampler: tp.Optional[base.Optimizer] = None
        self._no_hypervolume = self._config.multiobjective_adaptation

    def set_objective_weights(self, weights: tp.Any) -> None:
        self.objective_weights = weights

    def recommend(self) -> p.Parameter:  # This is NOT the naive version. We deal with noise.
        sample_size = int((self.dimension * (self.dimension - 1)) / 2 + 2 * self.dimension + 1)
        if self._config.high_speed and len(self.archive) >= sample_size:
            try:
                meta_data = metamodel.learn_on_k_best(self.archive, sample_size)
                return self.parametrization.spawn_child().set_standardized_data(meta_data)
            except metamodel.MetaModelFailure:  # The optimum is at infinity. Shit happens.
                pass  # MetaModel failures are something which happens, no worries.
        if self._config.recommendation != "noisy":
            return self.current_bests[self._config.recommendation].parameter
        med_fitness = np.median([p.loss for p in self.population.values() if p.loss is not None])
        good_guys = [p for p in self.population.values() if p.loss is not None and p.loss < med_fitness]
        if not good_guys:
            return self.current_bests["pessimistic"].parameter
        data: tp.Any = sum(
            [g.get_standardized_data(reference=self.parametrization) for g in good_guys]
        ) / len(good_guys)
        out = self.parametrization.spawn_child()
        with p.helpers.deterministic_sampling(out):
            out.set_standardized_data(data)
        return out

    def _internal_ask_candidate(self) -> p.Parameter:
        if len(self.population) < self.llambda:  # initialization phase
            init = self._config.initialization
            if self.sampler is None and init == "QO":
                self.sampler = oneshot.SamplingSearch(
                    sampler="Hammersley", scrambled=True, opposition_mode="quasi"
                )(self.parametrization, budget=self.llambda)
            if self.sampler is None and init == "SO":
                self.sampler = oneshot.SamplingSearch(
                    sampler="Hammersley", scrambled=True, opposition_mode="special"
                )(self.parametrization, budget=self.llambda)
            if self.sampler is None and init not in ["gaussian", "parametrization"]:
                assert init in ["LHS", "QR"]
                self.sampler = oneshot.SamplingSearch(
                    sampler=init if init == "LHS" else "Hammersley", scrambled=init == "QR", scale=self.scale
                )(
                    self.parametrization,
                    budget=self.llambda,
                )
            if init == "parametrization":
                candidate = self.parametrization.sample()
            elif self.sampler is not None:
                candidate = self.sampler.ask()
            elif self._config.crossover == "voronoi":
                new_guy = (
                    self.scale * self._rng.normal(0, 1, self.dimension)
                    if len(self.population) > self.llambda / 6
                    else self.scale * self._rng.normal() * np.ones(self.dimension)
                )
                candidate = self.parametrization.spawn_child().set_standardized_data(new_guy)
            else:
                new_guy = self.scale * self._rng.normal(0, 1, self.dimension)
                candidate = self.parametrization.spawn_child().set_standardized_data(new_guy)
            candidate.heritage["lineage"] = candidate.uid  # new lineage
            self.population[candidate.uid] = candidate
            self._uid_queue.asked.add(candidate.uid)
            return candidate
        # init is done
        lineage = self._uid_queue.ask()
        parent = self.population[lineage]
        candidate = parent.spawn_child()
        candidate.heritage["lineage"] = lineage  # tell-not-asked may have provided a different lineage
        data = candidate.get_standardized_data(reference=self.parametrization)
        # define all the different parents
        uids = list(self.population)
        a, b = (self.population[uids[self._rng.randint(self.llambda)]] for _ in range(2))
        best = self.current_bests["pessimistic"].parameter
        # redefine the different parents in case of multiobjective optimization
        if self._config.multiobjective_adaptation and self.num_objectives > 1:
            pareto = self.pareto_front()
            # can't use choice directly on pareto, because parametrization can be iterable
            if pareto:
                best = parent if parent in pareto else pareto[self._rng.choice(len(pareto))]
            if len(pareto) > 2:  # otherwise, not enough diversity
                a, b = (pareto[idx] for idx in self._rng.choice(len(pareto), size=2, replace=False))
        # define donor
        data_a, data_b, data_best = (
            indiv.get_standardized_data(reference=self.parametrization) for indiv in (a, b, best)
        )
        donor = data + self._config.F1 * (data_a - data_b) + self._config.F2 * (data_best - data)
        candidate.parents_uids.extend([i.uid for i in (a, b)])
        # apply crossover
        co = self._config.crossover
        if co == "parametrization":
            candidate.recombine(self.parametrization.spawn_child().set_standardized_data(donor))
        else:
            crossovers = Crossover(
                self._rng, 1.0 / self.dimension if co == "dimension" else co, self.parametrization
            )
            crossovers.apply(donor, data)
            candidate.set_standardized_data(donor, reference=self.parametrization)
        return candidate

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: tp.FloatLoss) -> None:
        uid = candidate.heritage["lineage"]
        if uid not in self.population:  # parent was removed, revert to tell_not_asked
            self._internal_tell_not_asked(candidate, loss)
            return
        self._uid_queue.tell(uid)  # only add to queue if not a "tell_not_asked" (from a removed parent)
        parent = self.population[uid]
        mo_adapt = self._config.multiobjective_adaptation and self.num_objectives > 1
        if mo_adapt:
            if self.objective_weights is None:
                self.objective_weights = np.ones(self.num_objectives)
            else:
                assert len(self.objective_weights) == self.num_objectives
        mo_adapt &= candidate._losses is not None  # can happen with bad constraints
        if not mo_adapt and loss <= base._loss(parent):
            self.population[uid] = candidate
        elif mo_adapt and (
            parent._losses is None
            or np.average(candidate.losses < parent.losses, weights=self.objective_weights) > self._rng.rand()
        ):
            # multiobjective case, with adaptation,
            # randomly replaces the parent depending on the number of better losses
            self.population[uid] = candidate
        elif self._config.propagate_heritage and loss <= float("inf"):
            self.population[uid].heritage.update(candidate.heritage)

    def _internal_tell_not_asked(self, candidate: p.Parameter, loss: tp.FloatLoss) -> None:
        discardable: tp.Optional[str] = None
        if len(self.population) >= self.llambda:

            if self.num_objectives == 1:  # singleobjective: replace if better
                uid, worst = max(self.population.items(), key=lambda p: base._loss(p[1]))
                if loss < base._loss(worst):
                    discardable = uid
            else:  # multiobjective: replace if in pareto and some parents are not
                pareto_uids = {c.uid for c in self.pareto_front()}
                if candidate.uid in pareto_uids:
                    non_pareto_pop = {c.uid for c in self.population.values()} - pareto_uids
                    if non_pareto_pop:
                        nonpareto = {c.uid: c for c in self.population.values()}[list(non_pareto_pop)[0]]
                        discardable = nonpareto.heritage["lineage"]
        if discardable is not None:  # if we found a point to kick, kick it
            del self.population[discardable]
            self._uid_queue.discard(discardable)
        if len(self.population) < self.llambda:  # if there is space, add the new point
            self.population[candidate.uid] = candidate
            # this candidate lineage is not candidate.uid, but to avoid interfering with other optimizers (eg: PSO)
            # we should not update the lineage (and lineage of children must therefore be enforced manually)
            self._uid_queue.tell(candidate.uid)


# pylint: disable=too-many-arguments, too-many-instance-attributes
class DifferentialEvolution(base.ConfiguredOptimizer):
    """Differential evolution is typically used for continuous optimization.
    It uses differences between points in the population for doing mutations in fruitful directions;
    it is therefore a kind of covariance adaptation without any explicit covariance,
    making it super fast in high dimension. This class implements several variants of differential
    evolution, some of them adapted to genetic mutations as in
    `Holland’s work <https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)#Two-point_and_k-point_crossover>`_),
    (this combination is termed :code:`TwoPointsDE` in Nevergrad, corresponding to :code:`crossover="twopoints"`),
    or to the noisy setting (coined :code:`NoisyDE`, corresponding to :code:`recommendation="noisy"`).
    In that last case, the optimizer returns the mean of the individuals with fitness better than median,
    which might be stupid sometimes though.

    Default settings are CR =.5, F1=.8, F2=.8, curr-to-best, pop size is 30
    Initial population: pure random.

    Parameters
    ----------
    initialization: "parametrization", "LHS" or "QR" or "QO" or "SO"
        algorithm/distribution used for the initialization phase. If "parametrization", this uses the
        sample method of the parametrization.
    scale: float or str
        scale of random component of the updates
    recommendation: "pessimistic", "optimistic", "mean" or "noisy"
        choice of the criterion for the best point to recommend
    crossover: float or str
        crossover rate value, or strategy among:
        - "dimension": crossover rate of  1 / dimension,
        - "random": different random (uniform) crossover rate at each iteration
        - "onepoint": one point crossover
        - "twopoints": two points crossover
        - "rotated_twopoints": more genetic 2p cross-over
        - "parametrization": use the parametrization recombine method
    F1: float
        differential weight #1
    F2: float
        differential weight #2
    popsize: int, "standard", "dimension", "large"
        size of the population to use. "standard" is max(num_workers, 30), "dimension" max(num_workers, 30, dimension +1)
        and "large" max(num_workers, 30, 7 * dimension).
    multiobjective_adaptation: bool
        Automatically adapts to handle multiobjective case.  This is a very basic **experimental** version,
        activated by default because the non-multiobjective implementation is performing very badly.
    high_speed: bool
        Trying to make the optimization faster by a metamodel for the recommendation step.
    """

    def __init__(
        self,
        *,
        initialization: str = "parametrization",
        scale: tp.Union[str, float] = 1.0,
        recommendation: str = "optimistic",
        crossover: tp.Union[str, float] = 0.5,
        F1: float = 0.8,
        F2: float = 0.8,
        popsize: tp.Union[str, int] = "standard",
        propagate_heritage: bool = False,  # experimental
        multiobjective_adaptation: bool = True,
        high_speed: bool = False,
    ) -> None:
        super().__init__(_DE, locals(), as_config=True)
        assert recommendation in ["optimistic", "pessimistic", "noisy", "mean"]
        assert initialization in ["gaussian", "LHS", "QO", "SO", "QR", "parametrization"]

        assert isinstance(scale, float) or scale == "mini"
        if not isinstance(popsize, int):
            assert popsize in ["large", "dimension", "standard", "small"]
        assert isinstance(crossover, float) or crossover in [
            "onepoint",
            "twopoints",
            "rotated_twopoints",
            "dimension",
            "random",
            "parametrization",
            "voronoi",
        ]
        self.initialization = initialization
        self.scale = scale
        self.high_speed = high_speed
        self.recommendation = recommendation
        self.propagate_heritage = propagate_heritage
        self.F1 = F1
        self.F2 = F2
        self.crossover = crossover
        self.popsize = popsize
        self.multiobjective_adaptation = multiobjective_adaptation


DE = DifferentialEvolution().set_name("DE", register=True)
TwoPointsDE = DifferentialEvolution(crossover="twopoints").set_name("TwoPointsDE", register=True)
VoronoiDE = DifferentialEvolution(crossover="voronoi").set_name("VoronoiDE", register=True)
RotatedTwoPointsDE = DifferentialEvolution(crossover="rotated_twopoints").set_name(
    "RotatedTwoPointsDE", register=True
)

LhsDE = DifferentialEvolution(initialization="LHS").set_name("LhsDE", register=True)
QrDE = DifferentialEvolution(initialization="QR").set_name("QrDE", register=True)
QODE = DifferentialEvolution(initialization="QO").set_name("QODE", register=True)
SPQODE = DifferentialEvolution(initialization="QO", popsize="small").set_name("SPQODE", register=True)
QOTPDE = DifferentialEvolution(initialization="QO", crossover="twopoints").set_name("QOTPDE", register=True)
LQOTPDE = DifferentialEvolution(initialization="QO", scale=10.0, crossover="twopoints").set_name(
    "LQOTPDE", register=True
)
LQODE = DifferentialEvolution(initialization="QO", scale=10.0).set_name("LQODE", register=True)
SODE = DifferentialEvolution(initialization="SO").set_name("SODE", register=True)
NoisyDE = DifferentialEvolution(recommendation="noisy").set_name("NoisyDE", register=True)
AlmostRotationInvariantDE = DifferentialEvolution(crossover=0.9).set_name(
    "AlmostRotationInvariantDE", register=True
)
RotationInvariantDE = DifferentialEvolution(crossover=1.0, popsize="dimension").set_name(
    "RotationInvariantDE", register=True
)

# Excellent for hyperparameter tuning.
DiscreteDE = DifferentialEvolution(crossover="dimension").set_name("DiscreteDE", register=True)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import errors
from nevergrad.parametrization import core
from nevergrad import optimizers
from nevergrad import callbacks


class Constraint(core.Operator):
    """Operator for applying a constraint on a Parameter

    Parameters
    ----------
    func: function
        the constraint function, taking the same arguments that the function to optimize.
        This constraint function must return a float (or a list/tuple/array of floats),
        positive if the constraint is not satisfied, null or negative otherwise.
    optimizer: str
        name of the optimizer to use for solving the constraint
    budget: int
        the budget to use for applying the constraint

    Example
    -------
    >>> constrained_parameter = Constraint(constraint_function)(parameter)
    >>> constrained_parameter.value  # value after trying to satisfy the constraint
    """

    _LAYER_LEVEL = core.Level.CONSTRAINT

    def __init__(self, func: tp.Callable[..., tp.Loss], optimizer: str = "NGOpt", budget: int = 100) -> None:
        super().__init__()
        self._func = func
        self._opt_cls = optimizers.registry[optimizer]
        self._budget = budget
        self._cache: tp.Any = None

    def _layered_del_value(self) -> None:
        self._cache = None  # clear cache!

    def apply_constraint(self, parameter: core.Parameter) -> core.Parameter:
        """Find a new parameter that better satisfies the constraint"""
        # This function can be overriden
        optim = self._opt_cls(parameter, budget=self._budget)
        early_stopping = callbacks.EarlyStopping(self.stopping_criterion)
        optim.register_callback("ask", early_stopping)
        optim.minimize(self.function)
        return optim.pareto_front()[0]

    def function(self, *args: tp.Any, **kwargs: tp.Any) -> tp.Loss:
        out = self._func(*args, **kwargs)
        if isinstance(out, (bool, np.bool_)):
            raise errors.NevergradTypeError(
                "Constraint must be a positive float if unsatisfied constraint (not bool)"
            )
        return np.maximum(0, out)  # type: ignore

    def parameter(self) -> core.Parameter:
        """Returns a constraint-free parameter, for the optimization process"""
        param = self._layers[0].copy()
        # remove last layer and make sure it is the last one
        if self._layer_index != param._layers.pop()._layer_index:
            raise RuntimeError("Constraint layer should be unique and placed last")
        return param  # type: ignore

    def stopping_criterion(self, optimizer: tp.Any) -> bool:
        """Checks whether a solution was found
        This is used as stopping criterio callback
        """
        if optimizer.num_tell < 1:
            return False
        best = optimizer.pareto_front()[0]
        return not np.any(best.losses > 0)

    def _layered_get_value(self) -> tp.Any:
        # TODO: this can be made more efficient (fewer copy) if need be.
        # Override only apply_constraint if you can, tampering with this method is tricky
        if self._cache is not None:
            return self._cache
        parameter = self.parameter()
        satisfied = not np.any(self.function(*parameter.args, **parameter.kwargs))
        if satisfied:
            self._cache = parameter.value
            return self._cache
        root: core.Parameter = self._layers[0]  # type: ignore
        recom = self.apply_constraint(parameter)
        root.set_standardized_data(np.zeros(root.dimension), reference=recom)
        self._cache = recom.value
        return self._cache



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from . import cycling


def test_cycling() -> None:
    func = cycling.Cycling(30)
    x = 0 * np.random.rand(func.dimension)
    assert func(x)
    np.testing.assert_almost_equal(func(x), 282.8, decimal=1)
    for index in [31, 22, 23, 45, 61]:
        func2 = cycling.Cycling(index)
        strategy = func2.parametrization.sample().value
        func2(strategy)



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .teampursuit import teampursuit
from .cyclist import Cyclist
from .simulationresult import simulationresult
import math


class womensteampursuit(teampursuit):

    team_size = 3
    race_distance = 3000
    lap_distance = 250
    race_segments = int((2 * (race_distance / lap_distance)) - 1)
    maximum_transitions = race_segments - 1

    def __init__(self):
        super().__init__()
        self.team = []
        for i in range(0, self.team_size):
            self.team.append(Cyclist(1.60, 65.0, 5.0, self, i + 1, "female"))

    def simulate(self, transition_strategy, pacing_strategy):

        if len(transition_strategy) != self.maximum_transitions:
            raise ValueError(
                f"Transition strategy for the womens team pursuit must have exactly {self.maximum_transitions} elements"
            )
        if len(pacing_strategy) != self.race_segments:
            raise ValueError(
                f"Pacing strategy for the womens team pursuit must have exactly {self.race_segments} elements"
            )
        for i in range(0, self.race_segments):
            if pacing_strategy[i] > Cyclist.max_power or pacing_strategy[i] < Cyclist.min_power:
                raise ValueError(
                    f"All power elements of the pacing strategy must be in the range {Cyclist.min_power}-{Cyclist.max_power} Watts"
                )

        for i in range(0, len(self.team)):
            self.team[i].reset()

        velocity_profile = [None] * self.race_segments
        proportion_completed = 0
        race_time = 0

        for i in range(0, self.race_segments):
            if i == 0 or i == (self.race_segments - 1):
                distance = 187.5
            else:
                distance = 125.0
            if super().cyclists_remaining() == 3:

                if i >= 1 and transition_strategy[i - 1]:
                    super().transition()
                    race_time += teampursuit.transition_time

                leader = super().leader()
                time = 0.0
                distance_ridden = 0.0
                while distance_ridden < distance:
                    dist = leader.set_pace(pacing_strategy[i])

                    for j in range(0, len(self.team)):
                        if self.team[j].get_position() > 1:
                            self.team[j].follow(dist)

                    if distance_ridden + dist <= distance:
                        distance_ridden += dist
                    else:
                        distance_ridden = distance

                    time += self.time_step

                leader.increase_fatigue()
                for j in range(0, len(self.team)):
                    if self.team[j].get_position() > 1:
                        self.team[j].recover()

                if super().cyclists_remaining() >= 3:
                    velocity_profile[i] = distance / time
                    race_time += time
                    proportion_completed += distance / self.race_distance
                else:
                    race_time = math.inf
            else:
                race_time = math.inf
                break

        remaining_energies = []
        for i in range(0, len(self.team)):
            remaining_energies.append(self.team[i].get_remaining_energy())

        return simulationresult(race_time, proportion_completed, remaining_energies, velocity_profile)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np


class Transform:
    """Defines a unique random transformation (index selection, translation, and optionally rotation)
    which can be applied to a point
    """

    def __init__(
        self,
        indices: tp.List[int],
        translation_factor: float = 1,
        rotation: bool = False,
        random_state: tp.Optional[np.random.RandomState] = None,
        expo: float = 1.0,
    ) -> None:
        dim = len(indices)
        assert dim
        if random_state is None:
            random_state = np.random.RandomState(0)
            random_state.set_state(np.random.get_state())
        self.indices = np.asarray(indices)
        self.translation: np.ndarray = (random_state.normal(0, 1, dim) ** expo) * translation_factor
        self.rotation_matrix: tp.Optional[np.ndarray] = None
        if rotation:
            self.rotation_matrix = np.linalg.qr(random_state.normal(0, 1, size=(dim, dim)))[0]

    def __call__(self, x: np.ndarray) -> np.ndarray:
        y: np.ndarray = x[self.indices] - self.translation
        if self.rotation_matrix is not None:
            y = self.rotation_matrix.dot(y)  # type: ignore
        return y



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Based on a discussion at Dagstuhl's seminar on Computational Intelligence in Games with:
# - Dan Ashlock
# - Chiara Sironi
# - Guenter Rudolph
# - Jialin Liu

import matplotlib.pyplot as plt
import numpy as np
from nevergrad.parametrization import parameter as p
from ..base import ExperimentFunction


class STSP(ExperimentFunction):
    def __init__(self, dimension: int = 500, complex_tsp: bool = False) -> None:
        super().__init__(self._simulate_stsp, p.Array(shape=(dimension,)))
        self.order = np.arange(0, self.dimension)
        self.complex = complex_tsp
        self.x = self.parametrization.random_state.normal(size=self.dimension)
        self.y = self.parametrization.random_state.normal(size=self.dimension)

    def _simulate_stsp(self, x: np.ndarray) -> float:
        order = np.argsort(x)
        self.order = order
        x = self.x[order]
        y = self.y[order]
        output = np.sqrt((x[0] - x[-1]) ** 2 + (y[0] - y[-1]) ** 2) + sum(
            np.sqrt((x[i] - x[i + 1]) ** 2 + (y[i] - y[i + 1]) ** 2) for i in range(self.dimension - 1)
        )
        if self.complex:
            output += np.sum(np.diff(order) ** 2) / np.sqrt(self.dimension + 1.0)
        return float(output)

    def make_plots(self, filename: str = "stsp.png") -> None:
        plt.clf()
        # Plot the optimization run.
        ax = plt.subplot(1, 1, 1)
        ax.set_xlabel("iteration number")
        order = self.order
        x = self.x
        y = self.y
        ax.plot((x[o] for o in order), (y[o] for o in order))
        plt.savefig(filename)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# This code is based on a code and ideas by Emmanuel Centeno and Antoine Moreau,
# University Clermont Auvergne, CNRS, SIGMA Clermont, Institut Pascal

import typing as tp
from math import pi, cos, sin
import matplotlib.pyplot as plt
import numpy as np
from nevergrad.parametrization import parameter as p
from ..base import ExperimentFunction


class Agent:
    """An agent has an input size, an output size, a number of layers, a width of its internal layers
    (a.k.a number of neurons per hidden layer)."""

    def __init__(self, input_size: int, output_size: int, layers: int = 3, layer_width: int = 14) -> None:
        assert layers >= 2
        self.input_size = input_size
        self.output_size = output_size
        self.layers = [np.zeros((layer_width, input_size))]
        for _ in range(layers - 2):
            self.layers += [np.zeros((layer_width, layer_width))]
        self.layers += [np.zeros((output_size, layer_width))]
        assert len(self.layers) == layers

    @property
    def dimension(self) -> int:
        return sum(layer.size for layer in self.layers)

    def set_parameters(self, weights: np.ndarray) -> None:
        if weights.size != self.dimension:
            raise ValueError(f"length = {weights.size} instead of {self.dimension}: {weights}.")
        start = 0
        for i, layer in enumerate(self.layers):
            numel = layer.size
            self.layers[i] = weights[start : start + numel].reshape(layer.shape)
            start += numel
        if start != weights.size:
            raise RuntimeError("Unexpected runtime error when distributing the weights")

    def get_output(self, data: np.ndarray) -> np.ndarray:
        for l in self.layers[:-1]:
            data = np.tanh(l @ data)
        return self.layers[-1] @ data  # type: ignore


# pylint: disable=too-many-instance-attributes,too-many-arguments,too-many-statements,too-many-locals
class PowerSystem(ExperimentFunction):
    """Very simple model of a power system.
    Real life is more complicated!

    Parameters
    ----------
    num_dams: int
        number of dams to be managed
    depth: int
        number of layers in the neural networks
    width: int
        number of neurons per hidden layer
    year_to_day_ratio: float = 2.
        Ratio between std of consumption in the year and std of consumption in the day.
    constant_to_year_ratio: float
        Ratio between constant baseline consumption and std of consumption in the year.
    back_to_normal: float
        Part of the variability which is forgotten at each time step.
    consumption_noise: float
        Instantaneous variability.
    num_thermal_plants: int
        Number of thermal plants.
    num_years: float
        Number of years.
    failure_cost: float
        Cost of not satisfying the demand. Equivalent to an expensive infinite capacity thermal plant.
    """

    def __init__(
        self,
        num_dams: int = 13,
        depth: int = 3,
        width: int = 3,
        year_to_day_ratio: float = 2.0,
        constant_to_year_ratio: float = 1.0,
        back_to_normal: float = 0.5,
        consumption_noise: float = 0.1,
        num_thermal_plants: int = 7,
        num_years: float = 1.0,
        failure_cost: float = 500.0,
    ) -> None:
        self.num_dams = num_dams
        self.losses: tp.List[float] = []
        self.marginal_costs: tp.List[float] = []
        # Parameters describing the problem.
        self.year_to_day_ratio = year_to_day_ratio
        self.constant_to_year_ratio = constant_to_year_ratio
        self.back_to_normal = back_to_normal
        self.consumption_noise = consumption_noise
        self.num_thermal_plants = num_thermal_plants
        self.number_of_years = num_years
        self.failure_cost = failure_cost
        self.hydro_prod_per_time_step: tp.List[tp.Any] = []  # TODO @oteytaud initial values?
        self.consumption_per_time_step: tp.List[tp.Any] = []

        self.average_consumption = self.constant_to_year_ratio * self.year_to_day_ratio
        self.thermal_power_capacity = self.average_consumption * np.random.rand(self.num_thermal_plants)
        self.thermal_power_prices = np.random.rand(num_thermal_plants)
        self.dam_agents = [
            Agent(10 + num_dams + 2 * self.num_thermal_plants, depth, width) for _ in range(num_dams)
        ]
        parameter = p.Instrumentation(
            *[p.Array(shape=(int(a.dimension),)) for a in self.dam_agents]
        ).set_name("")
        super().__init__(self._simulate_power_system, parameter)
        self.parametrization.function.deterministic = False

    def get_num_vars(self) -> tp.List[tp.Any]:
        return [m.dimension for m in self.dam_agents]

    def _simulate_power_system(self, *arrays: np.ndarray) -> float:
        failure_cost = (
            self.failure_cost
        )  # Cost of power demand which is not satisfied (equivalent to a expensive infinite thermal group).
        dam_agents = self.dam_agents
        for agent, array in zip(dam_agents, arrays):
            agent.set_parameters(array)

        self.marginal_costs = []

        num_dams = int(self.num_dams)
        # Assume empty initial stocks.
        stocks = np.zeros((num_dams,))
        # Nonsense delays.
        delay = np.cos(np.arange(num_dams))
        cost = 0.0
        # Loop on time steps.
        num_time_steps = int(365 * 24 * self.number_of_years)
        consumption = 0.0
        hydro_prod_per_time_step: tp.List[tp.Any] = []
        consumption_per_time_step: tp.List[float] = []
        for t in range(num_time_steps):

            # Rain
            stocks += 0.5 * (1.0 + np.cos(2 * pi * t / (24 * 365) + delay)) * np.random.rand(num_dams)
            # Consumption model.
            base_consumption = (
                self.constant_to_year_ratio * self.year_to_day_ratio
                + 0.5 * self.year_to_day_ratio * (1.0 + cos(2 * pi * t / (24 * 365)))
                + 0.5 * (1.0 + cos(2 * pi * t / 24))
            )

            if t == 0:
                consumption = base_consumption
            else:
                consumption = max(
                    0.0,
                    consumption
                    + self.consumption_noise * (np.random.rand() - 0.5)
                    + self.back_to_normal * (base_consumption - consumption),
                )
            consumption_per_time_step += [consumption]
            # "Needed" stores what we need, and will decrease as we use various power plants for producing.
            needed = consumption

            # Setting inputs for all agents.
            base_x = [
                cos(2 * pi * t / 24.0),
                sin(2 * pi * t / 24.0),
                cos(2 * pi * t / (365 * 24)),
                sin(2 * pi * t / (365 * 24)),
                needed,
                self.average_consumption,
                self.year_to_day_ratio,
                self.constant_to_year_ratio,
                self.back_to_normal,
                self.consumption_noise,
            ]

            x = np.concatenate((base_x, self.thermal_power_capacity, self.thermal_power_prices, stocks))  # type: ignore

            # Prices as a decomposition tool!
            price: np.ndarray = np.asarray([a.get_output(x)[0] for a in dam_agents])
            dam_index = np.arange(num_dams)
            price = np.concatenate((price, self.thermal_power_prices))
            capacity = np.concatenate((stocks, self.thermal_power_capacity))
            dam_index = np.concatenate((dam_index, -1 * np.ones(len(price), dtype=int)))

            assert len(price) == num_dams + self.num_thermal_plants
            hydro_prod: np.ndarray = np.zeros(num_dams)

            # Let us rank power plants by production cost.
            order = sorted(range(len(price)), key=lambda x: price[x])  # pylint: disable=cell-var-from-loop
            price = price[order]
            capacity = capacity[order]
            dam_index = dam_index[order]

            # Using power plants in their cost order, so that we use cheap power plants first.
            marginal_cost = 0.0
            for i, _ in enumerate(price):
                if needed <= 0:
                    break
                production = min(capacity[i], needed)
                # If this is a dam, producing will reduce the stock.
                if dam_index[i] >= 0:
                    hydro_prod[dam_index[i]] += production  # Let us log the hydro prod for this dam.
                    stocks[dam_index[i]] -= production
                    assert stocks[dam_index[i]] >= -1e-7
                else:
                    # If this is not a dam, we pay for using thermal plants.
                    cost += production * price[i]
                    if production > 1e-7:
                        marginal_cost = price[i]

                needed -= production
            # Cost in case of failures -- this is
            # harming industries and hospitals, so it can be penalized.
            cost += failure_cost * needed
            if needed > 1e-7:
                marginal_cost = failure_cost
            self.marginal_costs += [marginal_cost]

            hydro_prod_per_time_step += [hydro_prod]
        # Other data of interest: , hydro_prod, hydro_prod_per_time_step, consumption_per_time_step
        assert len(hydro_prod_per_time_step) == num_time_steps  # Each time steps has 1 value per dam.
        assert len(consumption_per_time_step) == num_time_steps
        self.hydro_prod_per_time_step = hydro_prod_per_time_step
        self.consumption_per_time_step = consumption_per_time_step
        self.losses += [cost]
        return cost

    def make_plots(self, filename: str = "ps.png") -> None:
        losses = self.losses
        num_dams = self.num_dams
        consumption_per_ts = self.consumption_per_time_step
        hydro_prod_per_ts = self.hydro_prod_per_time_step
        total_hydro_prod_per_ts = [sum(h) for h in hydro_prod_per_ts]

        # Utility function for plotting per year or per day.
        def block(x: tp.List[float]) -> tp.List[float]:
            result: tp.List[float] = []
            step = int(np.sqrt(len(x)))
            for i in range(0, len(x), step):
                result += [sum(x[i : i + step]) / len(x[i : i + step])]
            return result

        def block24(x: tp.List[float]) -> tp.List[float]:
            result: tp.List[float] = []
            for i in range(0, len(x), 24):
                result += [sum(x[i : i + 24]) / len(x[i : i + 24])]
            if len(x) != len(result) * 24:
                print(len(x), len(result) * 24)
            return result

        def deblock24(x: tp.List[float]) -> tp.List[float]:
            result = [0.0] * 24
            for i, _ in enumerate(x):
                result[i % 24] += x[i] / 24.0
            return result

        plt.clf()
        # Plot the optimization run.
        ax = plt.subplot(2, 2, 1)
        ax.set_xlabel("iteration number")
        smoothed_losses = block(losses)
        ax.plot(np.linspace(0, 1, len(losses)), losses, label="losses")
        ax.plot(np.linspace(0, 1, len(smoothed_losses)), smoothed_losses, label="smoothed losses")
        ax.legend(loc="best")

        # Plotting marginal costs.
        ax = plt.subplot(2, 2, 4)
        marginal_cost_per_hour = deblock24(self.marginal_costs)
        marginal_cost_per_day = block24(self.marginal_costs)
        ax.plot(
            np.linspace(0, 0.5, len(marginal_cost_per_hour)),
            marginal_cost_per_hour,
            label="marginal cost per hour",
        )
        ax.plot(
            np.linspace(0.5, 1, len(marginal_cost_per_day)),
            marginal_cost_per_day,
            label="marginal cost per day",
        )
        ax.legend(loc="best")

        # Plot consumption per day and decomposition of production.
        ax = plt.subplot(2, 2, 2)
        consumption_per_day = block24(consumption_per_ts)
        hydro_prod_per_day = block24(total_hydro_prod_per_ts)
        ax.plot(np.linspace(1, 365, len(consumption_per_day)), consumption_per_day, label="consumption")
        ax.plot(np.linspace(1, 365, len(hydro_prod_per_day)), hydro_prod_per_day, label="hydro")
        for i in range(min(num_dams, 3)):
            hydro_ts = [h[i] for h in hydro_prod_per_ts]

            hydro_day = block24(hydro_ts)
            ax.plot(np.linspace(1, 365, len(hydro_day)), hydro_day, label="dam " + str(i) + " prod")
        ax.set_xlabel("time step")
        ax.set_ylabel("production per day")
        ax.legend(loc="best")

        # Plot consumption per hour of the day and decomposition of production.
        ax = plt.subplot(2, 2, 3)
        consumption_per_hour = deblock24(consumption_per_ts)
        hydro_prod_per_hour = deblock24(total_hydro_prod_per_ts)
        ax.plot(np.linspace(1, 24, len(consumption_per_hour)), consumption_per_hour, label="consumption")
        ax.plot(np.linspace(1, 24, len(hydro_prod_per_hour)), hydro_prod_per_hour, label="hydro")
        for i in range(min(num_dams, 3)):
            hydro_ts = [h[i] for h in hydro_prod_per_ts]
            hydro_hour = deblock24(hydro_ts)
            ax.plot(np.linspace(1, 24, len(hydro_hour)), hydro_hour, label="dam " + str(i) + " prod")
        ax.set_xlabel("time step")
        ax.set_ylabel("production per hour")
        ax.legend(loc="best")
        plt.savefig(filename)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Based on https://github.com/Foloso/MixSimulator

from .. import base


class OptimizeMix(base.ExperimentFunction):
    """
    MixSimulator is an application with an optimization model for calculating
    and simulating the least cost of an energy mix under certain constraints.

    For now, it uses a default dataset (more will be added soon).

    For more information, visit : https://github.com/Foloso/MixSimulator

    Parameters
    ----------
    time: int
        total time over which it evaluates the mix (must be in hour)

    """

    def __init__(self, time: int = 8760) -> None:
        try:
            from mixsimulator.MixSimulator import MixSimulator  # pylint: disable=import-outside-toplevel
            from mixsimulator.Demand import Demand

            self._mix = MixSimulator()
            self._mix.set_data_to("Toamasina")
            self._demand = Demand()
            self._demand.set_data_to("Toamasina", delimiter=",")
            self._mix.set_demand(self._demand)

        except (KeyError, AttributeError, ModuleNotFoundError) as e:
            # send a skip error so that this does not break the test suit
            raise base.UnsupportedExperiment("mixsimulator dependency issue") from e
        self._mix.set_penalisation_cost(100)
        self._mix.set_carbon_cost(10)
        parameters = self._mix.get_opt_params(time)
        parameters.set_name("dims")
        super().__init__(self._mix.loss_function, parameters)
